package PureMessage::Manager::Response::Reports;

###############################################################################
# Required inclusions.
###############################################################################
use strict;
use PureMessage::Manager::Utils qw(urlize);
use ActiveState::DateTime qw(month_name_short);
use HTML::Template;
use URI::Escape;
use base qw(PureMessage::Manager::Response);

###############################################################################
# Default values.
###############################################################################
my %defaults = (
    'height'    => 250,
    'width'     => 600,
    );

our %config;

###############################################################################
# Creates a new reports response object.
###############################################################################
sub new
{
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = $class->SUPER::new( @_ );
    # override charset default
    $self->{'html_meta_charset'} = 'utf-8';
    $self->add_taskmenu(
        { 'title' => 'Report Categories',
          'buttons' => [[
            { 'type' => 'text', 'text' => 'Reports Summary', 'href' => 'index.cgi' },
            { 'type' => 'text', 'text' => 'Reports',         'href' => 'index.cgi?v=list' },
            ]],
        } );
    eval {
        require PureMessage::Reports::Report;
        require PureMessage::Reports::Report::Config;
        PureMessage::Reports::Report::Config->import(qw(%config));
        require PureMessage::Manager::Response::Reports::Utils;
        PureMessage::Manager::Response::Reports::Utils->import(qw(_mk_html_chart REPORTS_CGI));
    };
    if ($@) {
        $self->raise_error("Error: $@");
        $self->send();
        return;
    }
    return $self;
}

###############################################################################
# Shows the "reports summary" dashboard page.
###############################################################################
sub show_dashboard
{
    my ($self) = @_;

    # XXX load from a resource file
    $self->title( 'Summary' );
    $self->addDashboard( 'module' => $self->module() );
}

###############################################################################
# Shows a list of built-in reports.
###############################################################################
sub show_report_list
{
    my ($self) = @_;

    # XXX load from a resource file
    $self->title( 'Reports' );
    $self->_show_reports( 'tmpl/reportlist.html', $config{'report'} );
}

###############################################################################
# INTERNAL METHOD.  Shows the given list of reports using the given template.
###############################################################################
sub _show_reports
{
    my ($self, $tmplfile, $reporthash) = @_;

    ###########################################################################
    # Get the info on all of the provided reports.
    ###########################################################################
    my @reports;
    foreach my $rpt (keys %{$reporthash})
    {
        my $rpt_info = $reporthash->{$rpt};
        my $format   = ${$rpt_info->{'format'}}[0];
        #######################################################################
        # NOTE: 12h has been hardcoded as the default choice for the history in
        # a couple places ... so it is being added here as well ... just to
        # keep things consistent.  Uggggggly, but oh well.
        #######################################################################
        my $tmplinfo = {
            'id'    => $rpt,
            'title' => $rpt_info->{'title'},
            'desc'  => $rpt_info->{'desc'},
            'icon'  => 'img/reports_' . $format . '.gif',
            'href'  => "index.cgi?v=report&history=12h&id=$rpt&format=$format",
            };
        push( @reports, $tmplinfo );
    }

    ###########################################################################
    # Sort the report list by ID.
    ###########################################################################
    @reports = sort { $a->{'id'} cmp $b->{'id'} } @reports;

    ###########################################################################
    # Add the template to show the list of reports.
    ###########################################################################
    my $template = new HTML::Template( filename => $tmplfile,
                                       die_on_bad_params => 0,
                                     );
    $template->param( reports => \@reports );
    $self->add_part( $template, 'body_part' )
}

###############################################################################
# Shows a single report.
###############################################################################
sub show_report
{
    my ($self, %args) = @_;

    ###########################################################################
    # Use defaults where other values not provided.
    ###########################################################################
    map { $args{$_} = $defaults{$_} if (!defined $args{$_}) } keys %defaults;


    ###########################################################################
    # Get the report object.
    ###########################################################################
    my $report = new PureMessage::Reports::Report( $args{'id'} );
    $self->title( $report->get_title() );

    if((!exists($args{per_host})) && exists($report->{'_report'}->{per_host})) {
        $args{per_host} = $report->{'_report'}->{per_host};
    }
    
    ###########################################################################
    # Based on the requested format, either generate the HTML for the report
    # table, -or- generate HTML to graph the report.
    ###########################################################################
    my $html = '';
    if ($args{'format'} eq 'tbl') {
        $html = $report->run_report( %args );
    }
    else {
        $html = _mk_html_chart( $report, %args );
    }
    $self->add_part( $html );

    ###########################################################################
    # Make the report nav.
    ###########################################################################
    my $showlegend = $report->has_data() && ($args{'format'} ne 'tbl');
    $self->_mk_report_nav(
            %args,
            ($showlegend ? ('legend' => $report->get_legend()) : ()),
            );
}

###############################################################################
# INTERNAL METHOD.  Makes the report navbar.
###############################################################################
sub _mk_report_nav
{
    my ($self, %opts) = @_;

    ###########################################################################
    # Add in our "Tasks" menu
    ###########################################################################
    my $report = new PureMessage::Reports::Report( $opts{'id'} );

    unless ($opts{'no_taskmenu'}) {
	# By default for export format & graph display we always include the 
        # start_<unit> and end_<unit> parameters ie start_hour, end_hour.
        my $def_params;
        foreach my $unit (qw(year mon day hour)) {
	    foreach my $tref (qw(start end)) {
		my $key = "$tref".'_'."$unit";
		$def_params .= "&$key=";
                $def_params .= uri_escape($opts{$key}) if exists $opts{$key};
	    }
	}

        $self->add_taskmenu(
            { 'title'   => 'Tasks',
              'buttons' => [[
                { 'type' => 'export',
                  'text' => 'Export',
                  'href' => 'chart.cgi?v=report&id=' .
                      urlize($opts{'id'}) .
                      '&format=csv' .
                      '&export=1' .
                      '&history=' .
                      urlize($opts{'history'}) .
		      $def_params,
                },
                { 'type' => 'schedule',
                  'text' => 'Schedule',
                  'href' => 'schedule.cgi?v=report&id=' .
                      urlize($opts{'id'}) .
                      '&title=' .
                      uri_escape($report->get_title()) .
                      '&history=' .
                      urlize($opts{'history'}) .
                      '&format=' . 
                      urlize($opts{'format'}) .
		      $def_params,
                },
                ]],
            } );
    }

    ###########################################################################
    # Quick modify for reports.
    ###########################################################################
    unless ($opts{'no_reportmod'}) {
        # XXX This should be in a language file
        my %formats_full= ('tbl' => 'Table',
                           'bars' => 'Bar Graph',
                           'bars_stacked' => 'Stacked Bar Graph',
                           'lines' => 'Line Graph',
                           );

        my @periods = (
            [ 'N/A',    'Use Start/End Times'],
            [ '12h',    'Past 12 hours' ],
            [ '24h',    'Past 24 hours' ],
            [ '7d',     'Past 7 days' ],
            [ '30d',    'Past 30 days' ],
            [ '60d',    'Past 60 days' ],
            [ '90d',    'Past 90 days' ],
            [ '180d',   'Past 180 days' ],
            [ '365d',   'Past year' ],
            [ '730d',   'Past 2 years' ],
            );

        my @per_host = (
            [ '0',      'All Servers Combined' ],
            [ '1',      'Each Individual Server' ]
        );

        # Build up the menu data.
        my %menu_data = (
            'id'            => $opts{'id'},
            'reports_cgi'   => REPORTS_CGI(),
            'format'        => [],
            'history'       => [],
            'per_host'      => [],
            'start_year'    => [],
            'start_mon'     => [],
            'start_day'     => [],
            'start_hour'    => [],
            'start_min'     => [],
            'end_year'      => [],
            'end_mon'       => [],
            'end_day'       => [],
            'end_hour'      => [],
            'end_min'       => [],
            );

        my $formats = ${${$config{'report'}}{$opts{'id'}}}{'format'};
        foreach my $fmt (@$formats) {
            my $item = { 'val' => $fmt, 'name' => $formats_full{$fmt} };
            $item->{'selected'} = 'selected' if ($fmt eq $opts{'format'});
            push( @{$menu_data{'format'}}, $item );
        }
        foreach my $per (@periods)
        {
            my $item = { 'val' => $per->[0], 'name' => $per->[1] };
            ###################################################################
            # Set the history to default to 12 hours
            ###################################################################
            $opts{'history'} = $opts{'history'} ? $opts{'history'} : '12h';
            $item->{'selected'} = 'selected' if ($per->[0] eq $opts{'history'});
            push( @{$menu_data{'history'}}, $item );
        }
        foreach my $i (@per_host)
        {
            my $item = { 'val' => $i->[0], 'name' => $i->[1] };
            $item->{'selected'} = 'selected'
                if (($opts{'per_host'} || 0) == $i->[0]);
            push ( @{$menu_data{'per_host'}}, $item );    
        }

        # get current time
        my ($csec, $cmin, $chour, $cmday, $cmon, $cyear) = localtime();
        $cyear = $cyear + 1900;
	$cmon++;

        for (my $year = 2004; $year <= ($cyear + 10); $year++) {
            my $start_item = { 'year' => $year, 'name' => $year };
            if (defined $opts{'start_year'}) {
                if ($year == $opts{'start_year'}) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($year == $cyear) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'start_year'}}, $start_item);
            my $end_item = { 'year' => $year, 'name' => $year };
            if (defined $opts{'end_year'}) {
                if ($year == $opts{'end_year'}) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($year == $cyear) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'end_year'}}, $end_item);
        }

        for (my $mon = 1; $mon <= 12; $mon++) {
            my $start_item = { 'month' => $mon, 'name' => month_name_short($mon) };
            if (defined $opts{'start_mon'}) {
                if ($mon == $opts{'start_mon'}) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($mon == $cmon) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'start_month'}}, $start_item);
            my $end_item = { 'month' => $mon, 'name' => month_name_short($mon) };
            if (defined $opts{'end_mon'}) {
                if ($mon == $opts{'end_mon'}) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($mon == $cmon) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'end_month'}}, $end_item);
        }

        for (my $day = 0; $day <= 31; $day++) {
            my $start_item = { 'day' => $day, 'name' => $day };
            if (defined $opts{'start_day'}) {
                if ($day == $opts{'start_day'}) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($day == $cmday) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'start_day'}}, $start_item);
            my $end_item = { 'day' => $day, 'name' => $day };
            if (defined $opts{'end_day'}) {
                if ($day == $opts{'end_day'}) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($day == $cmday) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'end_day'}}, $end_item);
        }

        for (my $hour = 0; $hour <= 23; $hour++) {
            my $format_hour = sprintf("%2.2d:00", $hour);
            my $start_item = { 'hour' => $hour, 'name' => $format_hour };
            if (defined $opts{'start_hour'}) {
                if ($hour == $opts{'start_hour'}) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($hour == $cmday) {
                    $start_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'start_hour'}}, $start_item);
            my $end_item = { 'hour' => $hour, 'name' => $format_hour };
            if (defined $opts{'end_hour'}) {
                if ($hour == $opts{'end_hour'}) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            else {
                if ($hour == $cmday) {
                    $end_item->{'selected'} = 'selected';
                }
            }
            push ( @{$menu_data{'end_hour'}}, $end_item);
        }

        # Add the menu.
        $self->add_template_menu(
            'Modify Report',
            'file'  => 'tmpl/rptmod.html',
            'data'  => \%menu_data,
            );
    }

    ###########################################################################
    # Add in the report legend
    ###########################################################################
    if ($opts{'legend'}) {
        $self->add_taskmenu(
            { 'title'   => 'Legend',
              'buttons' => [[
                { 'type' => 'custom',
                  'text' => $opts{'legend'},
                },
                ]],
            } );
    }
}

###############################################################################
1;
