package PureMessage::Reports::Config::DB;

use strict;
use Carp;
use DBI;
use PureMessage::Config qw(%config);
use PureMessage::Database::Config;

# This module encapsulates routines for updating the DBMS repository of
# configuration parameters required for the reporting capability.
# The main reason for the existance of this module is the need for a
# separate namespace for the code that directly iteracts with the DBMS.

# Note about transaction processing in this module:
# The entire reports configuration data set is updated in a single
# transaction. Any error will result in changes being rolled back.

# Error handling: The routines that manipulate cfg data die at first
# occurrence of an error; routines that may be invoked from other packages
# issue warning, log an error and rollback the db session.

#XXX The interval<=>second count translation doesn't really belong to
# this module.

# Initializes DB session, if necessary. In case this module is used
# on-line (that is, outside of the stand-alone pmx-prd-config executable),
# the caller should pass the db handle to avoid overhead.
sub new {
    my ($class, %opts) = @_;
    my $db_handle = $opts{db_handle};
    my $logger = $opts{logger};
    my $self = {};
    unless ($db_handle) {
        my $reports_data_store = $config{pmx_prd_store} || 'pmx_quarantine';
        my $db_config = PureMessage::Database::Config->new();
        my $dsn = $db_config->store_value($reports_data_store, "dsn") ||
            $ENV{'PMX_DBI_SOURCE'};

        # NOTE: We ensure that all database handles open using exactly the same
        # connection arguments so that DBI handles can be shared among multiple
        # packages.  See the following docs for more details.
        # http://perl.apache.org/docs/1.0/guide/databases.html#Opening_Connections_With_Different_Parameters
        my $connect_method = $ENV{MOD_PERL} ? "connect" : "connect_cached";
        $db_handle = DBI->connect($dsn, "", "",
                                 {RaiseError => 0,
                                  PrintError => 0,
                                  AutoCommit => 0})
            or croak ("Could not connect to database with dsn \"$dsn\": $!");
        $db_handle->{RaiseError}  = 1; 
        $db_handle->{PrintError}  = 0; 
        $db_handle->{AutoCommit}  = 0; 
        $db_handle->{HandleError} = undef;
        $self->{do_disconnect} = 1;
    }
    $self->{db_handle} = $db_handle;
    $self->{logger} = $logger;
    bless($self, $class);
    return $self;
}

# terminates the db session only if it's our own
sub DESTROY {
    my $self = shift;
    $self->{db_handle}->disconnect() if ($self->{do_disconnect});
}

sub log {
    my ($self, $label, $text) = @_;
    if ($self->{logger}) {
        $self->{logger}->log($label, $text);
    } else {
	print STDERR "$label: $text\n";
    }
}

# Stores default configuration values (hard-coded here) into the db.
sub prd_cfg_db_store_defaults {
    my $db_handle = shift->{db_handle};
    eval {
        # invoke stored procedures that reset cfg data
        $db_handle->do("select del_prd_cfg_data()");
        $db_handle->do("select ins_static_prd_cfg_data()");
        $db_handle->do("select ins_transient_prd_cfg_data()");
        $db_handle->commit();
    };
    croak "Error while storing default configuration data: $@" if $@;
}

# Deletes (current) configuration from the db; does not commit.
sub prd_cfg_db_delete {
    my $db_handle = shift->{db_handle};
    eval {
        $db_handle->do("update prd_fact set is_collected = 0");
        $db_handle->do("update prd_applicable_dimension set is_collected = 0");
        $db_handle->do("delete from prd_nullified_dimension");
        $db_handle->do("delete from prd_reduction_action");
    };
    croak "Error while deleting configuration data: $@" if $@;
}

# Stores (new) configuration data into the db; does commit; returns the
# counts of facts and dimensions updated
sub prd_cfg_db_store {
    my $self = shift;
    my $db_handle = $self->{db_handle};
    my $config_href = shift;
    my $fact_count = 0;
    my $fact_href = $config_href->{fact};

    # update fact meta data
    foreach my $fact (keys %$fact_href) {
        my $one_fact_href = $fact_href->{$fact};
        next unless $one_fact_href->{is_collected};
        $self->_update_fact_contents($fact, $one_fact_href->{dimension});
        $self->_add_reduction_actns($fact, $one_fact_href->{reduction_action});
        $fact_count++;
    }

    # update dimension meta data
    my $dim_count = 0;
    my $dim_href = $config_href->{dimension};
    foreach my $dim (keys %$dim_href) {
        $self->_update_dim($dim, $dim_href->{$dim}->{purgeable},
                                 $dim_href->{$dim}->{match_pattern},
                                 $dim_href->{$dim}->{dont_match_pattern});
        $dim_count++;
    }
    $db_handle->commit();
    return ($fact_count, $dim_count);
};

# returns undef if something goes wrong or on invalid input
# may be called on-line
sub prd_cfg_db_fact_is_collected {
    my $self = shift;
    my $db_handle = $self->{db_handle};
    my $fact = shift;
    my $collected;

    eval {
        my $stmt = $db_handle->prepare(<<"EOT");
select is_collected from prd_fact where name = '$fact'
EOT
        $stmt->execute();
        ($collected) = $stmt->fetchrow_array();
    };
    if ($@) {
        $self->log("ERROR", "Error while retrieving fact data: $@");
        $db_handle->rollback();
    }
    return $collected;
}

# returns undef if something goes wrong or on invalid input
# may be called on-line
sub prd_cfg_db_dim_is_collected {
    my $self = shift;
    my $db_handle = $self->{db_handle};
    my $fact = shift;
    my $dim = shift;
    my $collected;
    eval {
        my $stmt = $db_handle->prepare(<<"EOT");
select is_collected from prd_applicable_dimension
where fact_name = '$fact' and dimension_name = '$dim'
EOT
        $stmt->execute();
        ($collected) = $stmt->fetchrow_array();
    };
    if ($@) {
        $self->log("ERROR", "Error while retrieving dimension data: $@");
        $db_handle->rollback();
    }
    return $collected;
}

# reeturns an array with (match, don't match) filters for the given dimension
# may be called on-line
sub prd_cfg_db_dim_filters {
    my $self = shift;
    my $db_handle = $self->{db_handle};
    my $dim = shift;
    my ($match, $dont_match);
    eval {
        my $stmt = $db_handle->prepare(<<"EOT");
select match_pattern, dont_match_pattern from prd_dimension
where name = '$dim'
EOT
        $stmt->execute();
        ($match, $dont_match) = $stmt->fetchrow_array();
    };
    if ($@) {
        $self->log("ERROR", "Error while retrieving filtering patterns: $@");
        $db_handle->rollback();
    }
    return ($match, $dont_match);
}

# prints the contents of reports configuration db to the std output
# (formatted as configuration file)
sub prd_cfg_db_dump {
    my $db_handle = shift->{db_handle};
    my $indent = "    ";
    my $fact_stmt = "select name from prd_fact where is_collected = 1";
    eval {
        my $prep_fact_stmt = $db_handle->prepare($fact_stmt);
        $prep_fact_stmt->execute();
        while (my ($fact) = $prep_fact_stmt->fetchrow_array()) {
            print "<fact $fact>\n";
            print $indent, "is_collected = 1\n";

            # get applicable dimensions
            my $coll_dim_stmt = <<"EOT";
select dimension_name from prd_applicable_dimension
where fact_name = '$fact' and is_collected = 1
EOT
            my $prep_coll_dim_stmt = $db_handle->prepare($coll_dim_stmt);
            $prep_coll_dim_stmt->execute();
            while (my ($dim) = $prep_coll_dim_stmt->fetchrow_array()) {
                print $indent, "dimension = $dim\n";
            }

            # get reduction actions
            my $actn_stmt = <<"EOT";
select action_name, action_type, age_threshold, recurrence, aggregation_period
from prd_reduction_action where fact_name = '$fact'
EOT
            my $prep_actn_stmt = $db_handle->prepare($actn_stmt);
            $prep_actn_stmt->execute();
            while (my($name, $type, $age_threshold,
                      $recurrence, $period)= $prep_actn_stmt->fetchrow_array()) {
                print $indent, "<reduction_action $name>\n";
                print $indent, $indent, "type = '$type'\n";
                print $indent, $indent,
                      "age_threshold = '$age_threshold seconds'\n";
                print $indent, $indent, "recurrence = '$recurrence'\n";
                print $indent, $indent,
                      "aggregation_period = '$period seconds'\n" if $period;
                print $indent, "</reduction_action>\n";
            }
            print "</fact>\n";
        }
    };
    croak "Error while reading fact data: $@" if $@;
     

    # get dimensions
    my $dim_stmt = <<"EOT";
select name, is_purgeable, match_pattern, dont_match_pattern
from prd_dimension
EOT
    eval {
        my $prep_dim_stmt = $db_handle->prepare($dim_stmt);
        $prep_dim_stmt->execute();
        while (my($name, $is_purgeable, $match_pattern,
                  $dont_match_pattern) = $prep_dim_stmt->fetchrow_array()) {
            print "<dimension $name>\n";
            print $indent, "is_purgeable = $is_purgeable\n";
            print $indent, "match_pattern = $match_pattern\n"
                if $match_pattern;
            print $indent, "dont_match_pattern = $dont_match_pattern\n"
                if $dont_match_pattern;
            print "</dimension>\n";
        }
    };
    croak "Error while reading dimension data: $@" if $@;
}

# Updates information on recorded (or collected) dimensional data for
# the given fact; does not commit.
sub _update_fact_contents {
    my ($self, $fact, $dim_ref) = @_;
    my $db_handle = $self->{db_handle};
    my $stmt = <<"EOT";
update prd_applicable_dimension set is_collected = 1
where fact_name = ? and dimension_name = ?
EOT
    eval {
        my $prep_stmt = $db_handle->prepare($stmt);
        foreach my $dim (@$dim_ref) {
            $prep_stmt->execute($fact, $dim);

            # exactly 1 row should be updated
            my $rows_affected = $prep_stmt->rows();
            if (1 != $rows_affected) {
                carp("$rows_affected rows updated for fact $fact, dimension $dim");
            }
        }
        $db_handle->do("update prd_fact set is_collected = 1 where name = '$fact'");
    };
    croak "Error while updating fact data: $@" if $@;
}

# Inserts the supplied reduction action info; returns the count of
# inserted actions.
sub _add_reduction_actns {
    my ($self, $fact, $actions_ref) = @_;
    my $db_handle = $self->{db_handle};
    my $count = 0;
    my $stmt = <<"EOT";
insert into prd_reduction_action (fact_name, action_name,
                                  action_type, age_threshold,
                                  recurrence, aggregation_period)
values (?, ?, ?, ?, ?, ?)
EOT
    eval {
        my $prep_stmt = $db_handle->prepare($stmt);
        foreach my $action (keys %$actions_ref) {
            my $aggregation_period = $actions_ref->{$action}->{aggregation_period};
            my $recurrence = $actions_ref->{$action}->{recurrence};
            my $age_threshold = $actions_ref->{$action}->{age_threshold};
            $aggregation_period = $self->_get_seconds($aggregation_period);
            $age_threshold = $self->_get_seconds($age_threshold);
            $prep_stmt->execute($fact, $action, $actions_ref->{$action}->{type},
                                $age_threshold, $recurrence, $aggregation_period);
            foreach my $nullified (@{$actions_ref->{$action}->{nullified_dimension}}){
                $db_handle->do(<<"EOT");
insert into prd_nullified_dimension
(fact_name, action_name, dimension_name) values
('$fact', '$action', '$nullified')
EOT
            }
            $count++;
        }
    };
    croak "Error while inserting reduction actions: $@" if $@;
    return $count;
}

# updates dimension meta data
sub _update_dim {

    # too many parameters, yes
    my ($self, $dim, $purgeable, $match_pattern, $dont_match_pattern) = @_;
    my $db_handle = $self->{db_handle};

    my $stmt = <<"EOT";
update prd_dimension set
    is_purgeable = ?, match_pattern = ?, dont_match_pattern = ?
where name = ?
EOT
    eval {
        my $prep_stmt = $db_handle->prepare($stmt);
        $prep_stmt->execute($purgeable, $match_pattern, $dont_match_pattern, $dim);
    };
    croak "Error while updating dimension data: $@" if $@;
}

# Converts a user readable time interval to a second count; returns the
# second count or input value unchanged (if not understood);
# XXX this doesn't really belong here
sub _get_seconds {
    my ($self, $interval) = @_;
    my ($count, $unit) = split /\s+/, $interval if $interval;
    return $interval unless ($count and $unit);

    my %seconds_per_unit = ('second' => 1, 'minute' => 60, 'hour' => 3600,
                            'day' => 3600*24, 'year' => (3600*24*365+6*3600));
    $unit = lc $unit;
    $unit =~ s/s$//;
    return $count * $seconds_per_unit{$unit} if exists $seconds_per_unit{$unit};
    return $interval;
}

1;
