package PureMessage::Reports::Custom;
use strict;
use warnings;
use base qw(PureMessage::Config::Validator Exporter);
use ActiveState::Config qw(:path);
use PureMessage::Config qw(expand_filename);
use PureMessage::Database::Config;
use PureMessage::Reports::DB qw{get_dbh};
use DBI;
our @EXPORT_OK = qw(add_report delete_report get_report_data update_report refresh_reports sanitize_key);

our $CONF_FILE_PATH = 'etc/reportsv2.conf';

sub add_report {
    my ($key, $values, $title) = @_;
    return "Missing Key" unless $key;
    return "Missing values" unless defined $values && scalar @{$values};
    return "Missing title" unless $title;

    my $dbh;
    eval { $dbh = get_dbh(); };
    return "Failed to connect to database" if $@ or !defined $dbh;

    if (! _add_report_perm($dbh, $key)) {
        $dbh->rollback;
        return "Failed to write to 'binary_permission_item' in database";
    }

    my $success = 0;
    eval { $success = _add_to_conf($key, \@{$values}, $title); };
    if ($@) {
        $dbh->rollback;
        return "Failed parsing reports config file";
    }
    if (!$success) {
        $dbh->rollback;
        return "Failed to write to reports config file";
    }

    $dbh->commit;
    return 0;
}

sub refresh_reports {
    my $dbh;
    eval { $dbh = get_dbh(); };
    return 0 if $@ or !defined $dbh;

    my $rpts = get_report_data();
    # dont early exit, just sync conf to db
    foreach my $id (keys %{$rpts}) {
        my $k = $rpts->{$id}{key};
        _add_report_perm($dbh, $k);
    }
    $dbh->commit;
    return 1;
}

sub update_report {
    my ($key, $values, $title) = @_;
    return _update_conf($key, \@{$values}, $title);
}

sub delete_report {
    my ($key) = @_;
    return 0 unless $key;

    my $p    = _parse_writable();
    my $root = $p->root;

    my $node = _get_node_by_key($root, $key) || return 0;
    my $id   = CF_REC_ID($node->groupclass);

    my $dbh;
    eval { $dbh = get_dbh(); };
    return 0 if $@ or !defined $dbh;

    if (defined $id &&
        _delete_report_data($dbh, $id, $key) &&
        _delete_report_conf($root, $id)) {

        $dbh->commit;
        $p->save;
    } else {
        $dbh->rollback;
        return 0;
    }
    return 1;
}

sub sanitize_key {
    my ($key) = @_;
    $key =~ s/[^[:word:]]+/_/;
    $key =~ s/\s+/_/;
    $key =~ s/=+/_/;
    $key =~ s/_+/_/g;
    $key =  lc $key;
    return $key;
}

sub _delete_report_data {
    my ($dbh, $id, $key) = @_;
    my $sql = "DELETE FROM binary_permission_item WHERE permission_item = ?";
    $dbh->do($sql, undef, "reports.custommarkhits.$key");
    $sql = "DELETE FROM binary_permission_item WHERE permission_item = ?";
    $dbh->do($sql, undef, "reports.custommarkhits.$key.export");
    $sql = "DELETE FROM prd_custom_mark_hits WHERE mark_key = ?";
    $dbh->do($sql, undef, $id);
    return 1;
}

sub _add_report_perm {
    my ($dbh, $key) = @_;
    my $insert;
    return 0 if _check_dup_report_perm($dbh, $key);
    $insert = $dbh->prepare("insert into binary_permission_item
                            (permission_item, default_permission_bits)
                             values
                            (?, ?)");
    return 0 unless $insert->execute("reports.custommarkhits.$key", 1);
    $insert = $dbh->prepare("insert into binary_permission_item
                            (permission_item, default_permission_bits)
                             values
                            (?, ?)");
    return 0 unless $insert->execute("reports.custommarkhits.$key.export", 1);
    return 1;
}

sub _check_dup_report_perm {
    my ($dbh, $key) = @_;
    my $sql = "SELECT permission_item FROM binary_permission_item " .
              "WHERE permission_item = ? " .
              "   OR permission_item = ?";
    my $row = $dbh->selectrow_hashref($sql, undef, "reports.custommarkhits.$key", "reports.custommarkhits.$key.export");
    return exists $row->{permission_item};
}

sub CF_REC_PATH { "custom.rpt" . shift }
sub CF_REC_ID { 
    my $name = shift;
    return $1 if $name && $name =~ m/rpt(\d+)/;
    return;
}

sub _get_node_by_key {
    my ($root, $key) = @_;

    my $group = path_node($root, "custom");

    my $iter = $group->children;
    while (my $node = $iter->next) {
        next unless CF_REC_ID($node->groupclass);

        my $attrs = $node->children;
        while (my $attr = $attrs->next) {
            return $node if $attr->name eq 'key' && $attr->value eq $key;
        }
    }
    return;
}

sub _delete_report_conf {
    my ($root, $id) = @_;

    if (path_node($root, CF_REC_PATH($id))) {
        path_node($root, CF_REC_PATH($id), delete => 1);
        return 1;
    }
    return 0;
}

sub _update_conf {
    my ($key, $values, $title) = @_;
    my $p = _parse_writable();
    my $root = $p->root();

    if (my $node = _get_node_by_key($root, $key)) {
        path_value ($node, "title", $title) if $title;
        path_value ($node, "values", $values)
            if ref($values) eq 'ARRAY' && @{$values};
        return $p->save() ? 0 : 1;
    }
    return 0;
}

sub _add_to_conf {
    my ($key, $values, $title) = @_;
    my $p = _parse_writable();
    my $root = $p->root();

    my $id = _cf_nextid($root);

    my $node = path_node($root, CF_REC_PATH($id), create => 'group');
    path_value ($node, "title",  $title);
    path_value ($node, "key",    $key);
    path_value ($node, "values", $values);

    return $p->save() ? 0 : 1;
}

sub _cf_nextid {
    my ($root, $id) = @_;

    unless (defined $id) {
        $id = path_value($root, "custom.next_id");
        $id = defined $id ? $id + 1 : 1;
    }

    path_value($root, "custom.next_id", $id);
    return $id;
}

sub get_report_data {
    my $id = shift;
    my $p = _parse_readonly();
    my $ret = {};
    my $cf = $p->as_hash();

    my $rpts = $cf->{custom} || {};
    delete $rpts->{next_id};

    foreach my $k (keys %{$rpts}) {
        # $p->as_hash() returns empty values as hash, convert to empty array
        my $val = [];
        if (!(ref($rpts->{$k}->{values}) eq "HASH")) {
            $val = $rpts->{$k}->{values};
        }
        my $nk = CF_REC_ID($k);
        $ret->{$nk} = {
            key      => $rpts->{$k}->{key},
            title    => $rpts->{$k}->{title},
            'values' => $val,
        };
    }

    if (defined $id) {
        my $r = $ret->{$id};
        return $r if defined $r;
        return {};
    }
    return $ret;
}

sub _parse_readonly {
    my $p = ActiveState::Config->new(strict => 1);
    PureMessage::Reports::Custom->validate($p);
    $p->parse_file(expand_filename($CONF_FILE_PATH));
    return $p;
}

sub _parse_writable {
    my $p = ActiveState::Config->new(strict => 1);
    PureMessage::Reports::Custom->validate($p);
    $p->parse_file(expand_filename($CONF_FILE_PATH), writable => 1);
    return $p;
}

# Used by the Resource Validator
sub validate {
    my ($pkg, $p) = @_;
    $p->typedecl('custom.next_id',     '#int');
    $p->typedecl('custom.rpt*.title',  '#str');
    $p->typedecl('custom.rpt*.key',    '#str');
    $p->typedecl('custom.rpt*.values', '#str');
}

1;
