package PureMessage::Reports::LogReader;
use strict;
use Carp;
use PureMessage::Reports::Config;
use PureMessage::Logger;
use PureMessage::Config qw(%config);
use PureMessage::MessageLog qw(message_log decode_value);
use PureMessage::LogReader;
use PureMessage::Version qw($PREFIX);
use ActiveState::Config qw(:path);
use MIME::Base64;

our $p = ActiveState::Config->new();

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %data = @_;
    # NOTE: The type must be defined and must be one of message, pmx or
    # blocklist corresponding to message_log, pmx_log and blocklist_log.
    #
    # Should more log types be needed, they must be added here.
    return 0 unless (($data{'type'} eq 'message') || 
                     ($data{'type'} eq 'pmx') ||
                     ($data{'type'} eq 'blocklist'));
    my $self = bless \%data, $class;
    $self->{'log'} = PureMessage::Logger->new('ident' => 'log_reader');
    $self->{'debug_level'} = 0;
    $self->{'debug_level'} = $config{'debug_level'} if defined $config{'debug_level'};
    return $self;
}

sub type        { shift->{'type'} }
sub log_file    { shift->{'log_file'} }
sub lr          { shift->{'lr'} }
sub debug_level { shift->{'debug_level'} }

sub errstr {
    my $self = shift;
    $self->{'errstr'} = $_[0] if defined $_[0];
    return $self->{'errstr'};
}

sub open {
    my $self = shift;
    $self->{'log_file'} = $_[0] if $_[0];
    if ((! defined $self->log_file) || ($self->log_file eq '')) {
        # open message_log
        if ($self->type eq 'message') {
            $self->{'log_file'} = message_log();
        }
        # open pmx_log
        elsif ($self->type eq 'pmx') {
            $self->{'log_file'} = $config{'log_to'};
        }
        # open blocklist_log
        elsif ($self->type eq 'blocklist') {
            $self->{'log_file'} = $config{'blocklist'}->{'Policy'}->{'blocklist_log'};
        }
    }
    croak "Can't open the message log. Missing file name."
        unless $self->log_file;
    eval q{
        $self->{lr} = PureMessage::LogReader->new(
            -logfile => $self->log_file,
            -resume_info => $self->retrieve_resume_info,
        );
    };
    die $@ if $@;
    $self->vvlog("Message log opened for reading");
    1;
}

# returns the next raw line of text, skipping over empty lines.
sub fetchrow { 
    my $self = shift;
    return $self->{lr}->readline;
}


# returns a decoded row of key/value pairs
sub fetchrow_hashref {
    my $self = shift;
    return $self->parse_row($self->fetchrow);
}

# Returns a decoded row of key/value pairs that match the RE
sub fetchrow_hashref_match{
    my ($self, $match) = @_;
    my $line = $self->fetchrow_match($match);
    return $self->parse_row($line);
}

# Returns the next raw line of text that matches the RE
# The caller is responsible for ensuring the RE is valid
sub fetchrow_match {
    my ($self, $match ) = @_;
    my $re = qr/$match/;
    my $line = '';
    while ($line = $self->{lr}->readline) {
        last if $line =~ $re;
    }
    chomp($line) if $line;
    return $line;
}

# NOTE: Since the log files differ in their layout, we need differing
#   parsers for each type of log file.  If a new type of log file is
#   added, a new parser must be added as well.

sub parse_row {
    my ($self, $row) = @_;
    if ($self->type eq 'message') {
        return $self->parse_message_log_row($row);
    }
    elsif ($self->type eq 'pmx') {
        return $self->parse_pmx_log_row($row);
    }
    elsif ($self->type eq 'blocklist') {
        return $self->parse_blocklist_log_row($row);
    }
    
    return 0;
}

sub parse_message_log_row {
    my ($self, $row) = @_;
    return undef unless ($row &&
                         ($row =~ /^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}/));
    my($ts, @fields) = split(' ', $row);
    my %row = ('time_stamp' => $ts);
    for (@fields) {
        my($k, $v) = split(/=/, $_, 2);
        $v = 1 unless defined $v;
        $v = decode_value($v) if $v && $v =~ /^\?/;
        $k = lc($k);
        if (defined $row{$k}) {
            if (ref($row{$k}) eq 'ARRAY') {
                push(@{$row{$k}}, $v);
            }
            else {
                $row{$k} = [$row{$k}, $v];
            }
        }
        else {
            $row{$k} = $v;
        }
    }
    return \%row;
}

sub parse_pmx_log_row {
    my ($self, $row) = @_;
    return undef unless ($row &&
                         ($row =~ /^\d+ \d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}/));
    my($pr, $ts, $pid, @fields) = split(' ', $row);
    my %row = ('time_stamp' => $ts);
    for (@fields) {
        my($k, $v) = split(/=/, $_, 2);
        $v = 1 unless defined $v;
        $k = lc($k);
        if (defined $row{$k}) {
            if (ref($row{$k}) eq 'ARRAY') {
                push(@{$row{$k}}, $v);
            }
            else {
                $row{$k} = [$row{$k}, $v];
            }
        }
        else {
            $row{$k} = $v;
        }
    }
    return \%row;
}

sub parse_blocklist_log_row {
    my ($self, $row) = @_;
    return undef unless ($row &&
                         ($row =~ /^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}/));
    
    my($ts, $ip, $action, $reason, $id) = split(' ', $row);
    
    my %row = ('time_stamp' => $ts);
    $row{'relay'} = $ip;
    $row{'action'} = lc($action);
    
    if ($id) {    
        (undef, $row{'id'}) = split('=', $id);
    }
    else {
        $row{'id'} = 'none';
    }
    
    return \%row;
}

# It is important that the last position in the log file is
# remembered correctly.  If it is not, some messages will be counted
# multiple times, or possibly not at all.
#
# The module PureMessage::LogUtils provides a method for doing this
# via the resume_info() method.
#
# This information will be stored in a "state" file written using
# ActiveState::Config.

sub resume_info {
    my $self = shift;
    return $self->{lr}->resume_info();
}

sub retrieve_resume_info {
    my $self = shift;
    croak("No ActiveState::Config object!") unless $p;
    # parse file and create a lock
    $p->parse_file($self->_state_file, ('writable' => 1, 'timeout' => 5));
    return decode_base64(path_value($p->root(), "resume_info"));
}

sub save_resume_info {
    my $self = shift;
    my $resume_info = shift;
    $resume_info = $resume_info ? $resume_info : $self->resume_info;
    $resume_info = encode_base64($resume_info, '');
    croak("Cannot set the state for the last log read!") unless $p;
    path_value($p->root(), "resume_info", $resume_info);
    $p->save();
}

sub exists_resume_info {
    my $self = shift;
    return 1 if -r $self->_state_file;
    return 0;
}

sub reset_resume_info {
    my $self = shift;
    my $c = ActiveState::Config->new();
    path_value($c->root(), "resume_info", "N/A");
    $c->saveto($self->_state_file);
}
    
sub log {
    my ($self, $level, $msg) = @_;
    $self->{'log'}->log($level, $msg);
}

sub vlog {
    my ($self, $msg) = @_;
    $self->log('DEBUG', $msg) if $self->debug_level > 5;
}

sub vvlog {
    my ($self, $msg) = @_;
    $self->log('DEBUG', $msg) if $self->debug_level > 9;
}

sub vvvlog {
    my ($self, $msg) = @_;
    $self->log('DEBUG', $msg) if $self->debug_level > 12;
}

sub _state_file {
    my $self = shift;

    my $state_file_dir = "$PREFIX/var/reports";
    return $state_file_dir . "/" . $self->type . "_log_state.conf";
}

1;
