package PureMessage::Reports::Mailer;

use base qw(Exporter);
use warnings;
use PureMessage::Config qw(%config);
use PureMessage::Reports::Report;
use PureMessage::Reports::Report::Config qw(list_report_ids get_report_info);
use PureMessage::Queue qw(queue_mail);
use MIME::Entity;
use HTML::Template;

our @EXPORT_OK = qw(list_reports send_reports);

###############################################################################
# Displays a list of available Report IDs to the user.
###############################################################################
sub list_reports
{
    my @rptids = list_report_ids();
    return @rptids;
}

###############################################################################
# E-mails reports.
###############################################################################
sub send_reports
{
    my $options = shift;

    ###########################################################################
    # Get required options, and check that they're there.
    ###########################################################################
    my $mailto   = delete $options->{'to'};
    my $mailfrom = delete $options->{'from'} || $config{'admin'};
    my %args;

    #######################################################################
    # Build up report options.
    #######################################################################
    if ($options->{'start'} && $options->{'end'}) {
        $args{'period_start'} = $options->{'start'};
        $args{'period_end'}   = $options->{'end'};
    }
    else {
        $args{'history'} = $options->{'time-frame'} || '24h';
    }
    $args{'per_host'} = $options->{'individual'};
    $args{'format'}   = 'csv' unless $options->{'html'};

    ###########################################################################
    # Run each of the reports, and gather all of the results.
    ###########################################################################
    my @reports;
    foreach my $rid (@{$options->{'id'}})
    {
        #######################################################################
        # Run the report.
        #######################################################################
        my $report = new PureMessage::Reports::Report( $rid );
        $report->run_report( %args );

        #######################################################################
        # Add this report's info to the list.
        #######################################################################
        my $info = {
            'title'     => $report->get_title(),
            'desc'      => $report->get_description(),
            'legend'    => $report->get_legend(),
            'format'    => $report->get_format(),
            'mimetype'  => $report->mime_type(),
            };

        $info->{'isimage'} = 1 if ($info->{'format'} !~ /^(tbl|csv)$/);

        my $rep_output = $report->output();
        my $hosts = $report->get_hosts();
        for (my $i=0; $i<scalar(@$rep_output); $i++) {
            my %attributes = %$info;
            $attributes{'output'} = $rep_output->[$i];
            if ($hosts) {
                $attributes{'hostname'} = $hosts->[$i] . ":";
            }
            if ($info->{'isimage'}) {
                $attributes{'cid'} = _mk_content_id();
                my ($ext) = ($info->{'mimetype'} =~ m|/(.*)|);
                $attributes{'filename'} = $attributes{'cid'} . '.' . $ext;
            }
            push (@{$info->{'report_list'}}, \%attributes);
        }
        push( @reports, $info );
    }

    ###########################################################################
    # Based on the e-mail format to use (text vs html), create an appropriate
    # MIME message and stuff the body of the message into it.
    ###########################################################################
    my %hdrs = (
        'To'        => $mailto,
        'From'      => $mailfrom,
        'Subject'   => $options->{'subject'} || $reports[0]->{'title'},
        );
    my $mimemsg;
    if ($options->{'html'}) {
        $mimemsg = MIME::Entity->build( %hdrs,
                        'Type'  => 'multipart/related',
                        );
        $mimemsg->attach( 'Type' => 'text/html',
                          'Data' => _mk_html_body(\@reports),
                        );
    }
    else {
        $mimemsg = MIME::Entity->build( %hdrs,
                        'Type' => 'text/plain',
                        'Data'  => _mk_text_body(\@reports),
                        );
    }

    ###########################################################################
    # Add any attachments that are necessary; any reports that had a
    # "Content-ID" generated so that they could be dropped in inline.
    ###########################################################################
    foreach my $rpt (@reports)
    {
        foreach my $sub_rpt (@{$rpt->{'report_list'}}) {
            if ($sub_rpt->{'cid'})
            {
                $mimemsg->attach( 'Id'          => $sub_rpt->{'cid'},
                                  'Data'        => $sub_rpt->{'output'},
                                  'Filename'    => $sub_rpt->{'filename'},
                                  'Type'        => $rpt->{'mimetype'},
                                  'Disposition' => 'attachment',
                                  );
            }
        }
    }
    $mimemsg->print() if ($options->{'debug'});

    ###########################################################################
    # Send the message along on its way.
    ###########################################################################
    my $env_from = _get_raw_addr( $mailfrom );
    my $env_to   = _get_raw_addr( $mailto );
    queue_mail( from => $env_from, to => $env_to, message => $mimemsg->stringify() );
}

###############################################################################
# Generates a message body as HTML, and returns the HTML back to the caller.
###############################################################################
sub _mk_html_body
{
    my ($reports) = @_;
    my $template = new HTML::Template( filehandle => \*DATA,
                                       die_on_bad_params => 0,
                                     );
    $template->param( 'reports' => $reports );
    return $template->output();
}

###############################################################################
# Generates a message body as TEXT, and returns the text back to the caller.
###############################################################################
sub _mk_text_body
{
    my ($reports) = @_;
    my $body = '';
    foreach my $rpt (@{$reports})
    {
        $body .= $rpt->{'title'} . "\n";
        $body .= $rpt->{'desc'} . "\n\n";
        foreach my $report (@{$rpt->{'report_list'}}) {
            $body .= $report->{'hostname'}."\n\n" if exists $report->{'hostname'};
            $body .= $report->{'output'}."\n";
        }
        $body .= '-'x70 . "\n\n";
    }
    return $body;
}

###############################################################################
# Returns the first raw e-mail address found in the given address(es).
###############################################################################
sub _get_raw_addr {
    my $addr = shift;
    $addr =~ s/.*<(.*)>.*/$1/;
    return $addr;
}

###############################################################################
# Generates a unique "Content-ID", and returns it back to the caller.
###############################################################################
my $cid_counter = 0;
sub _mk_content_id
{
    return sprintf( '%x-%03d', time(), $cid_counter++ );
}

1;

__DATA__
<html>
<head>
  <style>
    /* standard table styles */
    td.tableHeader {
      background-color: #E5E2DA;
      font-weight: bold;
    }
    td.tableRow {
      background-color: #F2F1ED;
    }
    td.tableRowAlt {
      background-color: #F5F5F4;
    }
  </style>
</head>
<body>
  <table border="0" cellpadding="2" cellspacing="0">
    <TMPL_LOOP NAME=REPORTS>
      <tr>
        <td colspan="2" class="tableHeader">
          <TMPL_VAR NAME=TITLE>
        </td>
      </tr>
      <tr>
        <td colspan="2">
          <TMPL_VAR NAME=DESC>
        </td>
      </tr>
      <tr>
        <td>
          &nbsp;
        </td>
      </tr>
      <TMPL_LOOP NAME=REPORT_LIST>
         <tr valign=top>
            <td>
               <b><TMPL_VAR NAME=HOSTNAME></B>
            </td>
         </tr>
         <TMPL_IF NAME=ISIMAGE>
           <tr valign=top>
              <td>
                <img src="cid:<TMPL_VAR NAME=CID>" alt="<TMPL_VAR NAME=TITLE>">
              </td>
              <td>
                <TMPL_VAR NAME=LEGEND>
              </td>
           </tr>
         <TMPL_ELSE>
           <tr>
              <td colspan="2">
                <TMPL_VAR NAME=OUTPUT>
              </td>
           </tr>
         </TMPL_IF>
      </TMPL_LOOP>
      <tr>
        <td>
          &nbsp;
        </td>
      </tr>
    </TMPL_LOOP>
  </table>
</body>
</html>
