package PureMessage::Reports::Reductor;
use strict;
use warnings;
use Carp;
use DBI;
use HTTP::Date qw(time2iso);
use PureMessage::Database::Config;
use PureMessage::Reports::DB qw(get_dsn);
use PureMessage::Config qw(%config);
use PureMessage::Logger;

# This module provides most of the functionality for the PMX reporting
# data reduction utility, specifically the routines to connect to the
# database, read data reduction actions and execute them.
#
# The routines in this module perform considerable amount of database
# interaction; any error in database interaction will be logged and
# printed on stderr.
#
# Transaction processing: each reduction action is in its own transaction.

# Identifier for locking
my $capability = "PRD Data Reductor";

# establishes db connection and reads cfg data including reduction actions
sub new {
    my $class = shift;
    my %options = @_;
    my $self = {};
    bless($self, $class);

    $self->{'recurrence'} = $options{'recurrence'} || 'hourly';
    $self->{'verbose'}    = $options{'verbose'} || 0;
    $self->{'vacuum'}     = $options{'vacuum'} || 0;
    $self->{'purge'}      = $options{'purge'} || 0;
    $self->{'time_now'}   = $options{'time-now'} || time;

    $self->{'logger'} = PureMessage::Logger->new('ident' => "$capability");
    $self->{'debug'} = $config{'debug_level'};

    $self->_connect();
    $self->_lock_acquire() or return undef;

    $self->_get_meta_data();
    $self->_get_actions();

    return $self;
}

sub _vprint {
    my $self = shift;
    my $message = shift;
    if ($self->{'verbose'}) {
        print $capability, ": $message\n";
    }
}

# establishes 2 db connections; one for data manipulation and one for
# exclusion of other instances (locking)
sub _connect {
    my $self = shift;
    $self->_vprint("Establishing database connections...");
    my $dsn = get_dsn();

    my ($prd_handle, $lock_handle);
    eval {
        $prd_handle = DBI->connect(
            $dsn, "", "", {RaiseError => 1, PrintError => 0, AutoCommit => 0});
        $lock_handle = DBI->connect(
            $dsn, "", "", {RaiseError => 1, PrintError => 0, AutoCommit => 0})
    };
    if ($@) {
        my $msg = "Unable to connect to database with dsn $dsn: $@";
        $self->{'logger'}->log("ERROR", $msg);
        croak($msg);
    }

    $self->{'prd_handle'} = $prd_handle;
    $self->{'lock_handle'} = $lock_handle;
}

# we're not commiting here => the row stays locked and that's what we
# need in order to exclude other instances of this utility
sub _lock_acquire {
    my $self = shift;
    $self->_vprint("Acquiring lock...");
    $self->{'lock_handle'}->do("set statement_timeout = 1000");
    my $stmt = "update lock set running = ? " .
               "where running = 0 and capability = ?";
    my $stmt_handle;
    my $success;
    eval {
        $stmt_handle = $self->{'lock_handle'}->prepare($stmt);
        $stmt_handle->execute($$, $capability);
        $success = $stmt_handle->rows()
    };
    if ($@) {
        my $msg = "Error while acquiring lock: $@";
        $self->{'logger'}->log("ERROR", $msg);
        $self->{'lock_handle'}->rollback();
        carp $msg;
        $success = 0;
    }

    return $success;
}

sub _lock_release {
    my $self = shift;
    $self->{'lock_handle'}->rollback();
}

# Creates a temporary table for every fact table. The temp tables are
# automatically removed after the session is disconnected, so we just
# create all of them although most of them won't be needed. Returns the
# count of temp tables created.
sub _create_temp_tables {
    my $self = shift;
    $self->_vprint("Creating temporary tables...");
    my $count = 0;
    my $temp_table_name_ref = {};
    eval {
        my $stmt = $self->{'prd_handle'}->prepare("select name from prd_fact");
        $stmt->execute();
        while (my ($name) = $stmt->fetchrow_array()) {
            my $temp_table_name = "prd_tmp_$name";
            my $create_stmt = <<"EOT";
create temporary table $temp_table_name as
select * from prd_$name where 1 = 0
EOT
            if ($self->{'debug'}) {
                $self->{'logger'}->log("DEBUG", "Creating temporary table $temp_table_name");
            }
            $self->_vprint("$create_stmt");
            $self->{'prd_handle'}->do($create_stmt);
            $count++;
            $temp_table_name_ref->{$name} = $temp_table_name;
        }
        $self->{'prd_handle'}->commit;
    };

    if ($@) {
        my $msg = "Error while creating temporary tables: $@";
        $self->{'prd_handle'}->rollback();
        $self->{'logger'}->log("ERROR", $msg);
        carp $msg;
    }

    $self->{'temp_table_name'} = $temp_table_name_ref;
    return $count;
}

# reads facts, dimensions and their relationhip from the db and
# stores 2 hashes into $self: facts->applicable dimensions and purgeable
# dimensions->referencing facts
sub _get_meta_data {
    my $self = shift;
    my %fact;
    my %purgeable_dim;
    eval {
        my $stmt = $self->{'prd_handle'}->prepare(<<"EOT");
select fact_name, dimension_name, is_purgeable
from prd_applicable_dimension, prd_dimension
where prd_applicable_dimension.dimension_name = prd_dimension.name
EOT
        $stmt->execute();
        while (my ($fact, $dimension, $purgeable) = $stmt->fetchrow_array()) {
            push @{$fact{$fact}}, $dimension;
            push @{$purgeable_dim{$dimension}}, $fact if $purgeable;
        }
    };

    if ($@) {
        my $msg = "Error while reading meta data: $@";
        $self->{'prd_handle'}->rollback();
        $self->{'logger'}->log("ERROR", $msg);
        carp $msg;
    }

    $self->{'facts'} = \%fact;
    $self->{'purgeable_dim'} = \%purgeable_dim;
    return scalar %fact;
}

# Reads reduction actions from the db into an array of hashes, stores
# the array into $self and returns the count of actions read.
sub _get_actions {
    my $self = shift;
    $self->_vprint("Reading reduction actions...");

    my @action;
    my $action_count = 0;

    eval {
        my $read_actns_stmt = $self->{'prd_handle'}->prepare( <<EOT
select fact_name, action_name, action_type,
       age_threshold, recurrence, aggregation_period
from prd_reduction_action 
EOT
        );
        my $read_nlfd_dims_stmt = $self->{'prd_handle'}->prepare( <<EOT
select dimension_name from prd_nullified_dimension
where fact_name = ? and action_name = ?
EOT
        );

        $read_actns_stmt->execute();
        while (my ($fact_name, $action_name, $action_type,
                   $age_threshold, $recurrence,
                   $aggregation_period) = $read_actns_stmt->fetchrow_array()) {

            # if this was an aggregation action perhaps it
            # involves nullification of some dimensional data
            my $nlfd_dim_ref = [];
            $read_nlfd_dims_stmt->execute($fact_name, $action_name);
            while (my ($dim) = $read_nlfd_dims_stmt->fetchrow_array()) {
                push @$nlfd_dim_ref, $dim;
            }

            push @action, { 'fact_name'          => $fact_name,
                            'action_name'        => $action_name,
                            'action_type'        => $action_type,
                            'age_threshold'      => $age_threshold,
                            'recurrence'         => $recurrence,
                            'aggregation_period' => $aggregation_period,
                            'nlfd_dim'           => $nlfd_dim_ref,
                          };
            $action_count++;
        }
    };

    if ($@) {
        my $msg = "Error while reading reduction actions: $@";
        $self->{'prd_handle'}->rollback();
        $self->{'logger'}->log("ERROR", $msg);
        carp $msg;
    }

    $self->{'actions'} = \@action;
    return $action_count;
}

sub _process_action {
    my $self = shift;
    my $action_ref = shift;
    my $action_name = $action_ref->{'action_name'};
    my $fact_name = $action_ref->{'fact_name'};
    $self->_vprint("Processing action $action_name on $fact_name\n"); 
    # get SQL stmts that we'll process
    my ($action_type, $aggr_function) = split /\//, $action_ref->{'action_type'};
    my @sql;
    if ('expiry' eq $action_type) {
        @sql = $self->_get_expiry_sql($action_ref);
    } elsif (('aggregation' eq $action_type) and ('sum' eq $aggr_function)) {
        @sql = $self->_get_aggr_sum_sql($action_ref);
    } elsif (('aggregation' eq $action_type) and ('avg' eq $aggr_function)) {
        @sql = $self->_get_aggr_avg_sql($action_ref);
    } else {
        $self->{'logger'}->log("WARNING",
                               "Unsupported reduction action: $action_type");
    }

    eval {
        foreach my $stmt (@sql) {
            if ($self->{'debug'}) {
                $self->{'logger'}->log("DEBUG", "Executing statement: $stmt");
            }
            my $prepared_stmt = $self->{'prd_handle'}->prepare($stmt);
            $prepared_stmt->execute();

            # any data manipulation on this fact table means it's "affected"
            # in the sense that it may need vacuuming later on...
	    my $rows_processed;
	    if ($rows_processed = $prepared_stmt->rows()) {
                ${$self->{'affected'}}{"prd_$fact_name"} = 1;
            }
            $self->_vprint("$stmt $rows_processed rows\n"); 
        }
        $self->{'prd_handle'}->commit();
    };
    if ($@) {
        my $msg = "Error while processing action " .
                  "$action_name on $fact_name: $@";
        $self->{'prd_handle'}->rollback();
        $self->{'logger'}->log("ERROR", $msg);
        carp $msg;
    }
}

sub _get_expiry_sql {
    my $self = shift;
    my $action_ref = shift;
    my $curr_time = $self->{'time_now'};
    my $threshold = time2iso($curr_time - $action_ref->{'age_threshold'});
    return <<"EOT";
delete from prd_$action_ref->{'fact_name'}
where period_id in (
    select id from prd_period
    where period_end < '$threshold'
)
EOT
}

# Builds and returns SQL for aggregation/sum actions.
# XXX Limitations:
# - Re-aggregation of already aggregated data may not work properly
# - The aggregation period has to match the recurrence of invocation
# IOW, if we want hourly aggregation, we have to invoke aggregation every hour.
sub _get_aggr_sum_sql {
    my $self = shift;
    my $action_ref = shift;
    my $fact_name = $action_ref->{'fact_name'};
    my $action_name = $action_ref->{'action_name'};
    my $temp_table_name = $self->{'temp_table_name'}->{$fact_name};
    my $curr_time = $self->{'time_now'};
    my @all_dims = @{$self->{'facts'}->{$fact_name}};
    my @nlfd_dims = @{$action_ref->{'nlfd_dim'}};
    my $column_list = join ", ", map {"${_}_id"} @all_dims;
    my @aggr_sql = ();
    my $spec_period_end  =time2iso($curr_time - $action_ref->{'age_threshold'});
    my $spec_period_start=time2iso($curr_time - $action_ref->{'age_threshold'}
                                              - $action_ref->{'aggregation_period'});

    # some data doesn't yield itself to aggregation by summation
    if ('pmx_runtime_fact' eq $fact_name) {
        $self->{'logger'}->log("WARNING",
                               "Summary aggregation not applicable for " .
                               "$fact_name data.");
        return @aggr_sql;
    }

    # XXX the retrieval of period ids is a loftover from the previous
    # implementation - should be removed
    my ($period_start, $period_end, $period_start_id) =
        $self->_get_start_period($spec_period_start, $action_ref->{aggregation_period});

    unless ($period_start and $period_end) {
        $self->{'logger'}->log("WARNING",
                               "Action $action_name on fact $fact_name: " .
                               "Unable to retrieve start or stop period");
        return @aggr_sql;
    }

    # now get the id for the new period, the one we're aggregating over
    my $period_id = $self->_get_period_id($period_start, $period_end,
                                          $action_ref->{'aggregation_period'});
    my $select_list = $column_list;
    $select_list =~ s/\bperiod_id\b/$period_id/;
    my $group_by_list = $column_list;
    $group_by_list =~ s/\bperiod_id\b,?\s?//;

    foreach my $dim (@nlfd_dims) {

        # nullified dimensions become nulls
        $select_list =~ s/\b\Q${dim}\E_id\b/null/;

        # nullified dimensions don't figure in the group by clause
        $group_by_list =~ s/\b\Q${dim}\E_id\b,?\s?//;
    }

    # it only makes sense to aggregate over the periods that are shorter
    # than the resulting aggregation period
    my $sql_where = <<"EOT";
exists (
    select 'x' from prd_period
    where prd_period.duration < $action_ref->{'aggregation_period'}
    and prd_period.id = prd_$fact_name.period_id
    and '$period_start' <= prd_period.period_start
    and prd_period.period_end <= '$period_end'
)
EOT
    
    # XXX Once we decide to remove the restriction from the header we'll
    # have to dedicate much more attention to the "group by" clause for
    # this statement.
    push @aggr_sql, <<"EOT";
insert into $temp_table_name ($column_list, tally)
select $select_list, sum(tally) from prd_$fact_name
where $sql_where 
group by $group_by_list
EOT
    
    push @aggr_sql, <<"EOT";
delete from prd_$fact_name
where $sql_where
EOT

    push @aggr_sql, <<"EOT";
insert into prd_$fact_name
select * from $temp_table_name
where period_id = $period_id
EOT

    return @aggr_sql;
}

# Builds and returns SQL for aggregation/avg actions.
# XXX Same limitations as for aggregation/sum actions apply - see above.
sub _get_aggr_avg_sql {
    my $self = shift;
    my $action_ref = shift;
    my $fact_name = 'pmx_runtime_fact';
    my $action_name = $action_ref->{'action_name'};
    my $temp_table_name = $self->{'temp_table_name'}->{$fact_name};
    my $curr_time = $self->{'time_now'};
    my @aggr_sql = ();
    my $spec_period_end  =time2iso($curr_time - $action_ref->{'age_threshold'});
    my $spec_period_start=time2iso($curr_time - $action_ref->{'age_threshold'}
                                              - $action_ref->{'aggregation_period'});

    unless ($fact_name eq $action_ref->{fact_name}) {
        $self->{'logger'}->log("WARNING",
                               "Aggregation by averaging supported only for " .
                               "$fact_name data.");
        return @aggr_sql;
    }

    # gets actual start and stop time of the period over which the
    # aggregation is being performed, as recorded in the database
    # XXX we're retrieving numeric period ids here but not using them
    my ($period_start, $period_end, $period_start_id) =
        $self->_get_start_period($spec_period_start, $action_ref->{aggregation_period});

    unless ($period_start and $period_end) {
        $self->{'logger'}->log("WARNING",
                               "Action $action_name on fact $fact_name: " .
                               "Unable to retrieve start or stop period");
        return @aggr_sql;
    }
    my $period_id = $self->_get_period_id($period_start, $period_end,
                                          $action_ref->{'aggregation_period'});

    # This action is applicable to only one type of fact, so we're not
    # going to attempt to generalize the SQL generation code...
    my $sql_where = <<"EOT";
exists (
    select 'x' from prd_period
    where prd_period.duration < $action_ref->{'aggregation_period'}
    and prd_period.id = prd_pmx_runtime_fact.period_id
    and '$period_start' <= prd_period.period_start
    and prd_period.period_end <= '$period_end'
)
EOT

    push @aggr_sql, <<"EOT";
insert into $temp_table_name (
    period_id, runtime_fact_type_id, pmx_host_id, value_relative, value_absolute
)
select $period_id, runtime_fact_type_id, pmx_host_id,
       avg(coalesce(value_relative, 0)), avg(coalesce(value_absolute, 0))
from prd_pmx_runtime_fact
where $sql_where 
group by runtime_fact_type_id, pmx_host_id
EOT

    push @aggr_sql, <<"EOT";
delete from prd_pmx_runtime_fact
where $sql_where
EOT

    push @aggr_sql, <<"EOT";
insert into prd_pmx_runtime_fact (
    period_id, runtime_fact_type_id, pmx_host_id,
    value_relative, value_absolute
)
select period_id, runtime_fact_type_id, pmx_host_id,
       value_relative, value_absolute
from $temp_table_name
where period_id = $period_id
EOT

    return @aggr_sql;
}

# returns the database id (surrogate key) of the specified period;
# the period will be inserted if necessary
sub _get_period_id {
    my $self = shift;
    my ($period_start, $period_end, $duration) = @_;


    if (not ($period_start && $period_end && $duration)) {
        return undef;
    }
    my $insert_sql = <<"EOT";
INSERT INTO prd_period (period_start, period_end, duration, super_period_start)
SELECT X.*
    FROM (
        SELECT ?::timestamptz AS period_start,
               ?::timestamptz AS period_end,
               ?::int         AS duration,
               ?::timestamptz AS super_period_start
    ) AS X
    LEFT JOIN prd_period USING (period_start, period_end)
WHERE id IS NULL
EOT

    my $retrieve_sql = <<"EOT";
select id from prd_period
where period_start = ? and period_end = ?
EOT

    my $id;
    my $stmt;
    eval {
        # first insert it if it doesn't exist, then retrieve it...
        $stmt = $self->{'prd_handle'}->prepare($insert_sql);
        $stmt->execute($period_start, $period_end, $duration, $period_start);
        $stmt->finish();
        $self->{'prd_handle'}->commit();
    };

    # unique constraint violation is not an error in this context
    if ( (not $@) or ($@ =~ /unique constraint/) ) {
        eval {
            $self->{'prd_handle'}->rollback();
            $stmt = $self->{'prd_handle'}->prepare($retrieve_sql);
            $stmt->execute($period_start, $period_end);
            ($id) = $stmt->fetchrow_array();
            $self->{'prd_handle'}->commit();
        };
    }

    if ($@) {
        $self->{'prd_handle'}->rollback();
        my $msg = "Error while retrieving period id: $@";
        $self->{'logger'}->log("ERROR", $msg);
        croak($msg);
    }

    return $id;
}

# Returns the (start,end,id) of the period of (duration) granularity
# containing $specified_period_start.
# NO LONGER DOES:
# If such a period doesn't exist then returns the id + the actual start time of the earliest
# period that starts after the specified time.

my %_duration2period = (3600 => 'hour', 86400 => 'day');
sub _get_start_period {
    my ($self, $specified_period_start, $duration) = @_;
    my ($actual_period_start, $actual_period_end, $period_id);
    my $period = $_duration2period{$duration} or die "Cannot handle duration=$duration";
    eval {
        my $stmt = $self->{'prd_handle'}->prepare(<<"__SQL__");
SELECT 	period_start, period_end, id
FROM	prd_period
WHERE	period_start = date_trunc('$period', '$specified_period_start'::timestamptz)
AND	duration = $duration
__SQL__
        $stmt->execute();
	my $results = $stmt->fetchrow_arrayref();
	if (ref($results) eq "ARRAY") {
            ($actual_period_start, $actual_period_end, $period_id) = 
            @$results;
	} else {	
	    #no data in the DB yet!	
	    return;
	}
    };

    if ($@) {
        $self->{'prd_handle'}->rollback();
        my $msg = "Error while retrieving start period id: $@";
        $self->{'logger'}->log("ERROR", $msg);
        croak($msg);
    }
    return ($actual_period_start, $actual_period_end, $period_id);
}

# iterate through actions and apply them, if their recurrence corresponds
# to the recurrence of the job that we're running
sub process_actions {
    my $self = shift;
    $self->_vprint("Processing reduction actions...");
    $self->_create_temp_tables() or die("Unable to create temporary tables.");
    foreach my $action_ref (@{$self->{'actions'}}) {
        if ($self->{'recurrence'} eq $action_ref->{'recurrence'}) {
            $self->_process_action($action_ref);
        }
    }
}

# Runs the XXX PGSQL-specific "vacuum" command on the specified tables.
# Returns the count of vacuumed tables.
sub vacuum_tables {
    my $self = shift;
    my $count = 0;
    my @affected_table = keys %{$self->{'affected'}};
    $self->_vprint("Vacuuming tables...");
    $self->{'prd_handle'}->{AutoCommit} = 1;
    while (my $table = shift @affected_table) {
        eval {$self->{'prd_handle'}->do("vacuum $table")};
        if ($@) {
            my $msg = "Error while vacuuming table $table";
            $self->{'logger'}->log("ERROR", $msg);
            carp $msg;
        } else {
            $count++;
        }
    }
    $self->{'prd_handle'}->{AutoCommit} = 0;
    return $count;
}

# "Purges" dimensional tables by removing all the rows that are not
# referenced by any of the fact data rows. Returns the count of purged
# tables.
sub purge_dimension_tables {
    my $self = shift;
    $self->_vprint("Purging unreferenced rows from dimension tables...");
    my $count = 0;
    foreach my $purgeable_dim (keys %{$self->{'purgeable_dim'}}) {
        my $sql = $self->_get_purge_sql(
            $purgeable_dim, @{${$self->{'purgeable_dim'}}{$purgeable_dim}});
        if ($self->{'debug'}) {
            $self->{'logger'}->log("DEBUG", "Executing statement: $sql");
        }
        my $prepared_stmt;
        eval {
            $prepared_stmt = $self->{'prd_handle'}->prepare($sql);
            $prepared_stmt->execute();
        };
        if ($@) {
            my $msg = "Error while executing statement: $sql";
            $self->{'prd_handle'}->rollback();
            $self->{'logger'}->log("ERROR", $msg);
            carp $msg;
        } else {
            $count++;
            if ($prepared_stmt->rows()) {
                ${$self->{'affected'}}{"prd_$purgeable_dim"} = 1;
            }
            $self->{'prd_handle'}->commit();
        }
    }
    return $count;
}

sub _get_purge_sql {
    my $self = shift;
    my $dim = shift;
    my @not_exists;

    # where clause is a conjunction of "not exists" predicates
    while (my $fact = shift) {
        my $not_exists = <<"EOT";
(not exists (select 1 from prd_$fact where ${dim}_id = prd_${dim}.id))
EOT
        push @not_exists, $not_exists;
    }
    my $where_cond = join " and ", @not_exists;

    # now put it together
    my $sql = "delete from prd_$dim where $where_cond";
    return $sql;
}

sub cleanup {
    my $self = shift;
    $self->{'prd_handle'}->disconnect() if $self->{'prd_handle'};
    $self->{'lock_handle'}->disconnect() if $self->{'lock_handle'};
}

sub DESTROY {
    my $self = shift;
    $self->cleanup();
}

1;

