package PureMessage::Reports::ReductorV2;
use strict;
use warnings;
use PureMessage::Version qw/$PREFIX/;
use PureMessage::Reports::DB qw(get_dbname);
use ActiveState::Run qw/run/;

# How many to retain in the top 'n' by-type reports
my $TOP_N = 20;
my $AUDIT = '--';	# $AUDIT = '' to show initial/reduced/pruned rowcounts.
sub new {
    my ($class, %opts) = @_;
    my $dbname = get_dbname() || 'pmx_quarantine';
    my $self = {
        'dbname' => $dbname,
    };
    bless $self, $class;
    return $self;
}

sub reduce {
    my $self = shift;
    my $sql = $self->reduce_sql;
    my $tmpfile = "$PREFIX/var/run/reduce.tmp.$$";
    END { unlink $tmpfile if $tmpfile }
    open(my $fh, ">$tmpfile") or die "Can't open $tmpfile: $!";
    print $fh $sql;
    close $fh or die "Can't write $tmpfile: $!";

    my $psql = "$PREFIX/postgres/bin/psql";
    die "Can't find psql at $psql" unless -x $psql;
    run("\@$psql -f $tmpfile $self->{dbname} > /dev/null");
    unlink $tmpfile;
}

sub reduce_sql {
    my $self = shift;
    return join("\n\n", <<EOT,
SELECT  max(period_start) AS omega INTO TEMP info FROM prd_msgs_by_type;
$AUDIT CREATE TEMP TABLE audit(name text, acount int, rcount int, pcount int);

-- Create a table to map time periods into the new range
SELECT * INTO TEMP tmp_period_map FROM (
        SELECT  period_start AS src_start,
                CASE WHEN period_start < omega - '31 days'::interval
                        THEN date_trunc('day', period_start)
                     ELSE date_trunc('hour', period_start)
                END AS tgt_start
        FROM    info,
                (SELECT distinct period_start
                 FROM prd_msgs_by_type, info
                 WHERE period_start < omega - '1hour'::interval
                 -- add   UNION ... prd_blocklist_stats when needed...
                ) AS Z
        ORDER BY 1
) AS X;

-- Delete targets that just map one period to itself.
DELETE FROM tmp_period_map
        WHERE   tgt_start NOT IN (
                SELECT tgt_start FROM tmp_period_map
                WHERE   tgt_start != src_start);

ANALYZE tmp_period_map;
ANALYZE info;
ANALYZE prd_state;
BEGIN;
EOT
	    _reduce_table('connections_by_action', 'action',              '1week'),
	    _reduce_table('msgs_by_policy_block',  'blocked',             '1week'),
	    _reduce_table('msgs_by_type',          'type,inbound',        '1week'),
	    _reduce_table('msgs_by_virus',         'virus_name,inbound',  '1hour'),
	    _reduce_table('policy_mark_hits',      'policy_mark',         '1week'),
	    _reduce_table('custom_mark_hits',      'mark_key,mark_val',   '1week'),
	    _reduce_table('relays_by_type',        'type,relay,inbound',  '1hour'),
	    _reduce_table('senders_by_type',       'type,sender,inbound', '1hour'),
	    _reduce_table('spam_by_recipient',     'recipient',           '1hour'),

	    _prune('msgs_by_virus',     'virus_name', 'inbound'),
	    _prune('relays_by_type',    'relay',      'type'),
	    _prune('senders_by_type',   'sender',     'type'),
	    _prune('spam_by_recipient', 'recipient',  ''),

            "UPDATE prd_state SET asof = (SELECT omega FROM info) WHERE name = 'rep-stable';",
	    "$AUDIT SELECT * FROM audit ORDER BY 1;\n",
            "COMMIT;\n"
        );
}

sub _reduce_table {
    my ($fact, $keys, $interval) = @_;
    my $allkeys = "group_name,period_start,$keys";
    return <<EOT;
-- REDUCE prd_$fact
$AUDIT INSERT INTO audit(name,acount) SELECT 'prd_$fact', COUNT(*) FROM prd_$fact;

SELECT  $allkeys, sum(tally) AS tally
INTO    tmp_$fact
FROM    prd_$fact
WHERE   period_start >= (SELECT asof FROM prd_state WHERE name = 'rep-stable')
GROUP BY $allkeys HAVING COUNT(*) > 1;

DELETE FROM prd_$fact 
WHERE ($allkeys) IN (SELECT $allkeys FROM tmp_$fact);

INSERT INTO tmp_$fact
SELECT group_name, tgt_start,$keys,SUM(tally)
FROM    prd_$fact
JOIN    tmp_period_map ON src_start = period_start
WHERE	tgt_start < (SELECT omega-'$interval'::interval FROM info)
GROUP BY tgt_start, group_name, $keys;

-- delete old periods
DELETE FROM prd_$fact
WHERE   period_start IN (
	SELECT src_start FROM tmp_period_map 
	WHERE   tgt_start < (SELECT omega-'$interval'::interval FROM info)
	);

-- insert new periods
INSERT INTO prd_$fact SELECT * FROM tmp_$fact;
$AUDIT UPDATE audit SET rcount = (SELECT COUNT(*) FROM prd_$fact) WHERE name = 'prd_$fact';

DROP TABLE tmp_$fact;
EOT
}

# 1) Remove data older than a week that has only a single reference
# 2) Prune any recipients below the median (tally) if that leaves
#    at least TOP_N rows per (period,type)

sub _prune {
    my ($fact, $col, $per) = @_;
    $per ||= '';
    my $table   = "prd_$fact";
    my $top     = 2*$TOP_N;
    my $cut_per	= '';
    my $and_per = '';
    if ($per) {
        $and_per = "AND $table.$per = CUT.$per";
	$cut_per = ",CUT.$per";
        $per = ",$per";
    }
    return <<EOT;
-- PRUNE $table
DELETE FROM $table
WHERE (period_start $per, $col) IN (
        SELECT date_trunc('hour',period_start) $per, $col
        FROM $table, info
        WHERE period_start < (SELECT omega FROM info) - '2 hour'::interval
        GROUP BY date_trunc('hour',period_start) $per, $col
        HAVING SUM(tally) = 1
      );

SELECT date_trunc('hour',period_start) AS period_start,group_name $per, SQRT(AVG(tally)/AVG(1.0/tally)) AS cut
INTO TEMP cut_$table
FROM $table
WHERE period_start < (SELECT omega FROM info) - '1 hour'::interval
GROUP BY date_trunc('hour',period_start),group_name $per
HAVING COUNT(*) > $top
AND SQRT(AVG(tally)/AVG(1.0/tally)) > MIN(tally);

CREATE UNIQUE INDEX cut_${table}_index ON cut_$table(period_start,group_name $per);

DELETE FROM $table
WHERE EXISTS(
        SELECT * FROM cut_$table AS CUT
        WHERE $table.period_start = CUT.period_start
	AND $table.group_name = CUT.group_name
        $and_per
        AND $table.tally        < CUT.cut
      );

$AUDIT UPDATE audit SET pcount = (SELECT COUNT(*) FROM prd_$fact) WHERE name = 'prd_$fact';
DROP TABLE cut_$table;

EOT
}

1;
