package PureMessage::Reports::Report;

###############################################################################
# Required inclusions.
###############################################################################
use strict;
use Data::Pivot qw(pivot);
use Time::Local;
use HTTP::Date qw(time2iso);
use ActiveState::DateTime qw(check_date);
use PureMessage::Logger;
use PureMessage::Reports::Report::Config qw(%reports_config);
use PureMessage::Reports::Report::Format qw(DCLRS);
use PureMessage::Reports::Config;
use PureMessage::Reports::DB qw(get_dbh);
use Carp qw(croak);

###############################################################################
# Default values.
###############################################################################
my %defaults = (
    ###########################################################################
    # Empty for now, add here any defaults if needed
    ###########################################################################
    );

###############################################################################
# Creates a new report object for the specified ReportID.
###############################################################################
sub new
{
    my ($proto, $reportid) = @_;

    ###########################################################################
    # Create a new object and bless it into our namespace.
    ###########################################################################
    my $class = ref($proto) || $proto;
    my $self = {};
    bless $self, $class;

    ###########################################################################
    # Get the report info.
    ###########################################################################
    $self->{'_report'} = $reports_config{'report'}{$reportid};
    die "Report $reportid was not found.\n" unless ($self->{'_report'});

    ###########################################################################
    # Grab the "labels" out of the report configuration.
    ###########################################################################
    if ($self->{'_report'}{'labels'})
    {
        my $labels = $self->{'_report'}{'labels'};
        if (ref($labels->{'label'}) eq 'ARRAY') {
            $self->{'_labels'} = [ @{$labels->{'label'}} ];
        }
        else {
            $self->{'_labels'} = [ $labels->{'label'} ];
        }
    }

    ###########################################################################
    # Create a logger.
    ###########################################################################
    $self->{'logger'} = PureMessage::Logger->new( 'ident' => 'reports' );

    ###########################################################################
    # Return the newly created object back to the caller.
    ###########################################################################
    return $self;
}

###############################################################################
# Helper methods to get info about the report (most of which are only useful
# -AFTER- running the report).
###############################################################################
sub get_title           { shift->{'_report'}{'title'} };
sub get_description     { shift->{'_report'}{'desc'} };
sub get_default_format  { ${shift->{'_report'}{'format'}}[0] };
sub get_format          { shift->{'_format'} };
sub get_labels          { shift->{'_labels'} };
sub get_hosts           { shift->{'_hosts'} };
sub get_data            { shift->{'_data'} || undef; }
sub has_data            { shift->{'_has_data'} };
sub output              { shift->{'_output'} };
sub mime_type           { shift->{'_mime_type'} };
sub needs_total         { shift->{'_report'}{'pivot_total'} }

# Unlike get_legend() that returns htmlified legend, this one only return a hash
# ref that compose of label as key and color as value;
sub get_legend_data
{
    my ($self, %opts) = @_;

    my $labels  = $self->get_labels();
    my @colours = DCLRS();

    my $legend = {};
    my $colours_size = scalar @colours;
    
    foreach my $idx (1 .. $#{$labels})
    {
        my $lbl = $labels->[$idx];
        # $#{$labels} might be greater than $#colours, so make sure that we wrap
        # around after the last index in @colours has been hit
        my $col = $colours[($idx-1) % $colours_size];

        $legend->{$lbl} = $col;
    }

    return $legend;
}

###############################################################################
# Returns a legend in HTML format, which should be called -AFTER- running the
# report.
#
# Accepts the same options as 'run_report()'.
###############################################################################
sub get_legend
{
    my ($self, %opts) = @_;

    ###########################################################################
    # Get the labels/colours that need to go into the legend.
    ###########################################################################
    my $labels  = $self->get_labels();
    my @colours = DCLRS();

    ###########################################################################
    # Generate the HTML legend and return it back to the caller.
    #
    # NOTE: We -always- start at offset "1" as the first element in the data
    #       set is going to be the X-axis and doesn't need to appear in the
    #       legend.
    ###########################################################################
    my @rows;
    my $colours_size = scalar @colours;
    
    foreach my $idx (1 .. $#{$labels})
    {
        my $lbl = $labels->[$idx];
        # $#{$labels} might be greater than $#colours, so make sure that we wrap
        # around after the last index in @colours has been hit
        my $col = $colours[($idx-1) % $colours_size];
        my $row = qq{
<tr>
  <td bgcolor="$col" style="border: 1px solid black">&nbsp;&nbsp;</td>
  <td nowrap="nowrap">$lbl</td>
</tr>
};
        push( @rows, $row );
    }

    my $html = '<table bgcolor="white" width="95%">' .
               join('',@rows) .
               '</table>';
    return $html;
}

###############################################################################
# Runs the report.  Results (in whatever format was specified) are returned to
# the caller as well as stored for later access via the 'output()' method.
#
# Accepted options (which over-ride any options present in the report config):
#   format          => format to produce report in (tbl, bars, etc)
#   period_start    => start time for report data (as epoch offset)
#   period_end      => end time for report data (as epoch offset)
#   history         => sets start/end time to "last 'n' hours/days"
#   grouping        => group data by "minute", "hour", "day", "month", or
#                      "year" if not provided, a suitable value is
#                      auto-selected based on the data range being analyzed
#   direction       => 'incoming' or 'outgoing' - limits report data
# ... and all report formatting options (from PM::Reports::Report::Format)
###############################################################################
sub _load_single_server_data
{
    my ($self, %opts) = @_;

    ###########################################################################
    # Fill in default values where they're not provided.
    ###########################################################################
    map { $opts{$_} = $defaults{$_} if (!defined $opts{$_}) } keys %defaults;

    ###########################################################################
    # Expand any provided "history" option, to determine what the actual start
    # and end of the period is.
    ###########################################################################
    if ($opts{'history'})
    {
        #######################################################################
        # Check to see if the history is set to N/A.  If it is set this means
        # that we should use the start and end times.  If not, just default to
        # using the history again.
        #######################################################################
        if ((defined $opts{'history'}) && ($opts{'history'} eq 'N/A')) {
            ###################################################################
            # Check for a valid date.
            ###################################################################
            unless (check_date($opts{'start_year'}, $opts{'start_mon'}, $opts{'start_day'})) {
                my $start_date = $opts{'start_year'} . "-" . $opts{'start_mon'} . "-" . $opts{'start_day'};
                die ("Invalid start date: $start_date.  Please re-enter the start date.");
            }
            unless (check_date($opts{'end_year'}, $opts{'end_mon'}, $opts{'end_day'})) {
                my $end_date = $opts{'end_year'} . "-" . $opts{'end_mon'} . "-" . $opts{'end_day'};
                die ("Invalid end date: $end_date.  Please re-enter the end date.");
            }
            ###################################################################
            # Convert to epochs
            ###################################################################
            $opts{'period_end'} = timelocal(0, 0, $opts{'end_hour'}, $opts{'end_day'}, ($opts{'end_mon'} - 1), ($opts{'end_year'} - 1900));
            $opts{'period_start'} = timelocal(0, 0, $opts{'start_hour'}, $opts{'start_day'}, ($opts{'start_mon'} - 1), ($opts{'start_year'} - 1900));
            ###################################################################
            # Check that start time is less than end time
            ###################################################################
            die ("Start time must be less than end time.")
                if ($opts{'period_start'} >= $opts{'period_end'});
        }
        else {
            my $now = time();
            $opts{'period_end'} = $now;
            $opts{'period_start'} = $now - _expand_duration($opts{'history'});
        }
    }

    ###########################################################################
    # Time diff is used to determine the grouping and the date format
    ###########################################################################
    my $t_diff = $opts{'period_end'} - $opts{'period_start'};

    ###########################################################################
    # Auto-select suitable grouping if none was provided either by the user or
    # by the report definition.
    ###########################################################################
    my $grouping = $opts{'grouping'} || $self->{'_report'}{'grouping'};
    if (!$grouping) {
        $grouping = 'minute';
        $grouping = 'hour'  if ($t_diff > _expand_duration('12h'));
        $grouping = 'day'   if ($t_diff > _expand_duration('7d'));
        $grouping = 'month' if ($t_diff > _expand_duration('180d'));
        $grouping = 'year'  if ($t_diff > _expand_duration('10y'));
        $opts{'grouping'} = $grouping;
        $self->{'logger'}->log( "DEBUG", "Auto-selected time-grouping '$grouping'." );
    }

    ###########################################################################
    # Date formatting will be done according to the duration of the report.
    # The formatting will be done at the SQL level, but this variable is used
    # to hold correct format before the SQL statement is build up.  These date
    # formats will be passed to the PostgreSQL function TO_CHAR.  See
    # PostgreSQL documentation for further details regarding TO_CHAR.
    ###########################################################################
    my $date_format = $opts{'date_format'} || $self->{'_report'}{'date_format'};
    if (!$date_format) {
        if ($t_diff <= _expand_duration('24h')) {
            $date_format = 'HH24:MI';
        }
        elsif (($t_diff > _expand_duration('24h')) && ($t_diff <= _expand_duration('1y'))) {
            $date_format = 'Mon DD HH24:MI';
        }
        else {
            $date_format = 'Mon DD YYYY HH24:MI';
        }
        $opts{'date_format'} = $date_format;
        $self->{'logger'}->log( "DEBUG", "Auto_selected date format '$date_format'." );
    }

    ###########################################################################
    # Convert any provided time periods from 'epoch' to 'ISO' formats.
    ###########################################################################
    map { $opts{$_} = time2iso($opts{$_}) if ($opts{$_}) } qw(period_start period_end);

    ###########################################################################
    # Generate the SQL to run the report for us.
    ###########################################################################
    my($sql, @values) = ($self->_generate_sql(%opts));

    ###########################################################################
    # Get the result set from the database.
    ###########################################################################
    my $dbh;
    eval { $dbh = get_dbh(); };
    if (!$dbh) {
        my $msg = "Unable to connect to report database";
        $self->{'logger'}->log( "ERROR", "$msg; $@" );
        die "$msg.\n";
    }
    $dbh->{RaiseError} = 0;     # Turn this off so we control show of errors

    # Disable PostgreSQL server-side prepares.
    # We are currently using invalid prepared SQL statements - since we're not
    # using named parameters, server-side prepares can't tell that the columns
    # in GROUP_BY are the same as the SELECTed columns.  This causes a
    # preparation error.
    $dbh->{pg_server_prepare} = 0;  

    my $sth = $dbh->prepare( $sql );
    if (!$sth) {
        my $msg = "Unable to prepare SQL to run report";
        $self->{'logger'}->log( "ERROR", "$msg; $DBI::errstr" );
        die "$msg.\n";
    }

    my $rc = $sth->execute(@values);
    if (!$rc) {
        my $msg = "Unable to execute SQL to run report";
        $self->{'logger'}->log( "ERROR", "$msg; $DBI::errstr" );
        die "$msg.\n";
    }

    my $results = $sth->fetchall_arrayref();
    $self->{'_has_data'} = scalar( @{$results} );

    $sth->finish();
    $dbh->disconnect();

    ###########################################################################
    # Figure out what format we're supposed to be generating the report in.
    ###########################################################################
    $opts{'format'} ||= $self->get_default_format();

    ###########################################################################
    # Pivot the results if necessary, optionally adding a "Total" column if the
    # report is in "table" or "csv" format and total option is specified in the
    # config file
    ###########################################################################
    if ($self->has_data())
    {
        # out of the pivoted reports, some require a 'Total' column, some don't
        my $dototal = 0;
        if (($opts{'format'} =~ /^(tbl|csv)$/) && $self->needs_total()) {
            $dototal = 1;
        }
        $results = $self->_pivot( $results, $dototal );
    }

    return $results;
}

sub run_single_report {
    my ($self, %opts) = @_;
    my $results = $self->_load_single_server_data(%opts);

    # XXX since we somehow split out load_data stuff and run_single_report
    # it appears that the fact that we use %opts as a copy wasn't taken into
    # consideration.
    $opts{format} ||= $self->get_default_format();

    my $output = $self->_format_results( $results, %opts );
    push (@{$self->{'_hosts'}}, $opts{'host_name'}) 
        if exists $opts{'host_name'};
    push (@{$self->{'_output'}}, $output);
    $self->{'_format'} = $opts{'format'};

    return $output;
}

sub load_data {
    my ($self, %opts) = @_;
    $self->{'_data'} = {};

    if($opts{per_host}) {
        my($hosts);
        my $dbh = get_dbh;
        if($hosts = $dbh->selectall_arrayref(
            "SELECT id, name FROM prd_pmx_host ORDER BY name",
            { Slice => {} }
        )) {
            my $i = 0;
            foreach my $host (@$hosts) {
                $i++;
                $self->{'_data'}->{$host->{host_name}} = $self->_load_single_server_data(
                    %opts, 
                    host_id => $host->{'id'},
                    host_name => $host->{'name'}
                ),
            }
        } else {
            die "Querying host list failed: ", $dbh->errstr;
        }
    } else {
        $self->{'_data'}->{'ALL'} = $self->_load_single_server_data(%opts);
    }
}

sub run_report {
    my($self, %opts) = @_;
    if($opts{per_host}) {
        my($hosts);
        my $dbh = get_dbh;
        if($hosts = $dbh->selectall_arrayref(
            "SELECT id, name FROM prd_pmx_host ORDER BY name",
            { Slice => {} }
        )) {
            my (@rv);
            my $i = 0;
            foreach my $host (@$hosts) {
                $i++;
                push(@rv,
                    qq|<p id="fqdn_host$i"><b>$host->{'name'}:</b></p>|,
                    $self->run_single_report(%opts, 
                                             host_id => $host->{'id'},
                                             host_name => $host->{'name'}),
                    "<br/>"
                );
            }
            return join("\n", @rv);
        } else {
            die "Querying host list failed: ", $dbh->errstr;
        }
    } else {
        return $self->run_single_report(%opts);
    }
}

###############################################################################
# List of dimensions and the actual fields in their respective tables that we
# want to use in the reports.
###############################################################################
my %dimension_fields = (
    'direction'             => 'name',
    'msg_size_range'        => ['lo', 'hi'],
    'period'                => 'period_start',
    'pmx_host'              => 'name',
    'pmx_process'           => 'name',
    'probability_range'     => ['lo', 'hi'],
    'quarantine_reason'     => 'name',
    'recipient'             => ['local', 'domain'],
    'relay'                 => 'address',
    'rsrc_utlztn_range'     => ['lo', 'hi'],
    'sender'                => ['local', 'domain'],
    'spam_virus_rule'       => 'rule',
    'policy_mark'           => 'name',
    'user_action'           => 'name',
    'runtime_fact_type'     => 'name',
    'atchmt_type'           => 'name',
    'mta_action'            => 'name',
    'policy_action'         => 'name'
    );

###############################################################################
# INTERNAL METHOD.  Generates the SQL to run the report.  Accepts the same
# options as the 'run_report()' method.
###############################################################################
sub _generate_sql
{
    my ($self, %opts) = @_;
    my $cfg = $self->{'_report'};

    ###########################################################################
    # Figure out which fact we're reporting on, and determine which table that
    # comes from.
    ###########################################################################
    my ($fact) = keys %{$cfg->{'fact'}};        # should only ever be one
    my $fact_info = $cfg->{'fact'}{$fact};
    my $tbl_fact = 'prd_' . $fact;

    ###########################################################################
    # Use a custom data point if given, otherwise default to 'tally'
    ###########################################################################
    my $data_pnt = $cfg->{'data_pnt'} ? $cfg->{'data_pnt'} : 'tally';

    ###########################################################################
    # Get the dimensions that we've been asked to gather, and those that we're
    # filtering on.
    ###########################################################################
    my @dimensions;
    if (ref($fact_info->{'dimension'}) eq 'ARRAY') {
        @dimensions = @{$fact_info->{'dimension'}};
    }
    else {
        @dimensions = ($fact_info->{'dimension'});
    }
    my @filtered = keys %{$cfg->{'dimension'}} if ($cfg->{'dimension'});
    my @orig_dimensions = @dimensions;

    ###########################################################################
    # Build the SQL...
    ###########################################################################
    my (@select, @from, @where, @group, @order, @values_group, @values_select,
        @values_where, @values_order, $limit) = ();

    ###########################################################################
    # Build up the SELECT clause...
    ###########################################################################
    #   1. dimensions
    my $range_order;
    foreach my $dim (@dimensions)
    {
        my $col = $self->_dimensional_column($dim);
        if ($dim eq 'period') {
            # Special case: group for time compression in output
            my $grouping  = $cfg->{'grouping'} || $opts{'grouping'};
            my $date_format = $cfg->{'date_format'} || $opts{'date_format'};
            $col = "TO_CHAR(DATE_TRUNC(?, $col), '$date_format')";
            push(@values_select, $grouping);
        }
        push( @select, $col );
    }
    #   2. tally (which may be aggregated)
    my $aggregate = _fact_aggregate( $fact );   # aggregate report results
    my $tally = 'prd_' . $fact . '.' . $data_pnt;
    $tally = "$aggregate($tally)" if $aggregate;
    push( @select, $tally );

    ###########################################################################
    # Make sure that we -always- include the "period" dimension for all other
    # clauses (if its not already in the list of dimensions to report on).
    ###########################################################################
    push( @dimensions, "period" ) if (!grep( /period/, @dimensions ));

    ###########################################################################
    # Build up the FROM clause...
    ###########################################################################
    #   1. fact table
    push( @from, $tbl_fact );
    #   2. dimensions
    map { push( @from, $self->_dimensional_table($_) ) } @dimensions;
    #   3. filtered dimensions
    map { push( @from, $self->_dimensional_table($_) ) } @filtered;

    ###########################################################################
    # Build up the WHERE clause...
    ###########################################################################

    #   0. host id
    if($opts{host_id}) {
        push(@where, "$tbl_fact\.pmx_host_id = ?");
        push(@values_where, $opts{host_id});
    }

    #   1. dimensional joins
    map {
        my $dim_tbl = $self->_dimensional_table( $_ );
	my $clean_field = _clean_field( $_ );
        
        # suppress the predicate if dimension coming from the fact table itself
        push(@where, "$tbl_fact\.$clean_field\_id=$dim_tbl\.id")
            unless ($tbl_fact eq $dim_tbl);

        } @dimensions;
    #   2. filtered dimensional joins
    map {
        my $dim_tbl = $self->_dimensional_table( $_ );
	my $clean_field = _clean_field( $_ );
        push(@where, "$tbl_fact\.$clean_field\_id=$dim_tbl\.id");
        } @filtered;
    #   3. filters
    foreach my $dim (@filtered)
    {
        my $field = $self->_dimensional_column( $dim );

        if ($cfg->{'dimension'}{$dim}{'match_filter_pattern'})
        {
            my $ptn = $cfg->{'dimension'}{$dim}{'match_filter_pattern'};
	    my @match_ptns = (ref($ptn) eq 'ARRAY') ? @{$ptn} : ($ptn);

	    my @sub_clauses;

            foreach (@match_ptns) {
                push(@sub_clauses, "$field ~* ?");
                push(@values_where, $_);
            }
                
	    my $clause = '(' . join( ' OR ', @sub_clauses ) . ')';
	    push( @where, $clause );
        }
        elsif ($cfg->{'dimension'}{$dim}{'dont_match_filter_pattern'})
        {
            my $ptn = $cfg->{'dimension'}{$dim}{'dont_match_filter_pattern'};
	    my @nomatch_ptns = (ref($ptn) eq 'ARRAY') ? @{$ptn} : ($ptn);

	    my @sub_clauses;

            foreach (@nomatch_ptns) {
                push(@sub_clauses, "$field !~* ?");
                push(@values_where, $_);
            }

	    my $clause = '(' . join( ' AND ', @sub_clauses ) . ')';
	    push( @where, $clause );
        }
    }
    #   4. time period
    my $p_start = $opts{'period_start'};
    my $p_end   = $opts{'period_end'};

    my $period_table = $self->_dimensional_table('period');

    if($p_start) {
        push(@where, "$period_table.period_start >= ?");
        push(@values_where, $p_start);
    }
    
    if($p_end) {
        push(@where, "$period_table.period_end <= ?");
        push(@values_where, $p_end);
    }

    # 5. direction
    my $direction = $opts{direction};
    if ($direction) {
        push @where, "direction = ?";
        push @values_where, $direction;
    }

    ###########################################################################
    # Build up the GROUP BY clause
    ###########################################################################
    if ($aggregate) {
        foreach my $dim (@orig_dimensions) {
            my $col = $self->_dimensional_column($dim);
            if ($dim eq 'period') {
                # Special case: group for time compression in output                                              
                my $grouping  = $cfg->{'grouping'} || $opts{'grouping'};
                my $date_format = $cfg->{'date_format'} || $opts{'date_format'};
                $col = "DATE_TRUNC(?, $col)";
                push(@values_group, $grouping);
            }
            push( @group, $col );
        }
    }

    ###########################################################################
    # Build up the ORDER BY clause
    ###########################################################################
    my $ord_info = $cfg->{'order'};
    if ($ord_info)
    {
        # Get the list of dimensions we're ordering on.
        my @ord_dimensions;
        if (ref($ord_info->{'dimension'}) eq 'ARRAY') {
            @ord_dimensions = @{$ord_info->{'dimension'}};
        }
        else {
            @ord_dimensions = ($ord_info->{'dimension'});
        }

        # Turn the list of ordered dimensions into SQL ordering.
        foreach my $dim (@ord_dimensions)
        {
            my ($dim_field, $ascdesc) = split( /:/, $dim );
            my $ord_field;
            if ($dim_field eq $data_pnt) {
                # Special case: order on the actual tally results
                $ord_field = $tally;
            }
            elsif ($dim_field eq 'period') {
                # Special case: group for time compression in output
                my $col = $self->_dimensional_column($dim_field);
                my $grouping  = $cfg->{'grouping'} || $opts{'grouping'};
                my $date_format = $cfg->{'date_format'} || $opts{'date_format'};
                $ord_field = "DATE_TRUNC(?, $col)";
                push(@values_order, $grouping);
            }
            elsif ($dim_field =~ /_range/) {
                # This casts the lower part of the range as an integer instead
                # of using it as a string
                my $col = $self->_dimensional_column($dim_field);
                $ord_field  = "CAST(substring(".$col.", 1, ";
                $ord_field .= "(position('-' in $col)-1)) AS integer)";
            }
            else {
                $ord_field = $self->_dimensional_column($dim_field);
            }
            # Add this field and its order to the SQL ordering
            $ascdesc ||= '';
            push( @order, join(' ', $ord_field, $ascdesc) );
        }
    } 

    ###########################################################################
    # Build up the LIMIT clause
    ###########################################################################
    $limit = "$opts{limit}" if $opts{limit};

    ###########################################################################
    # Uniq-ify the "FROM" and "WHERE" clauses
    ###########################################################################
    @from   = _uniqify_list( @from );
    _uniqify_placeholder_lists(\@where, \@values_where);

    ###########################################################################
    # Assemble the final SQL, returning it back to the caller.
    ###########################################################################
    my $sql = 'SELECT ' . join( ', ', @select ) . "\n" .
              '  FROM ' . join( ', ', @from ) . "\n" .
              ' WHERE ' . join( "\n   AND ", @where ) . "\n";
    $sql .= ' GROUP BY ' . join( ', ', @group ) . "\n" if (@group);
    $sql .= ' ORDER BY ' . join( ', ', @order ) . "\n" if (@order);
    $sql .= ' LIMIT ' . $limit . "\n" if ($limit);
    $self->{'logger'}->log(
        "DEBUG",
        "Generated report SQL:\n$sql\nbind values: '" .
            join("', '", @values_select, @values_where, @values_group, @values_order) .
            "'"
    );
    return($sql, @values_select, @values_where, @values_group, @values_order);
}

###############################################################################
# INTERNAL METHOD.  Returns the name of the table that the given dimension can
# be found in.
###############################################################################
sub _dimensional_table
{
    my ($self, $dim) = @_;
    ###########################################################################
    #
    # XXX - The hyphen workaround is a hack that should
    #        be fixed at some point ... it was used to
    #        override the default field selection as
    #        defined in this module
    #
    ###########################################################################
    if ($dim =~ /\-/) {
        ($dim) = split('-', $dim);
    }

    # period is special - if the fact table exposes it, we don't use
    # the dimensional table, we use the fact table itself; currently
    # it's only the facts that we retrieve through views that expose
    # the period details
    my $cfg = $self->{'_report'};
    my ($fact) = keys %{$cfg->{'fact'}};
    if (($fact =~ /_view\z/) and ($dim eq 'period')) {
        $dim = $fact;
    }

    return "prd_$dim";
}

###############################################################################
# INTERNAL METHOD.  Returns the full "table.column" name of the column which
# represents the given dimension.
#
# If a list-ref is provided, the value returned is the SQL equivalent of a
# join('-'...) of the appropriate columns.
###############################################################################
sub _dimensional_column
{
    my ($self, $dim) = @_;
    ###########################################################################
    #
    # XXX - The hyphen workaround is a hack that should
    #        be fixed at some point ... it was used to
    #        override the default field selection as
    #        defined in this module
    #
    ###########################################################################
    if ($dim =~ /\-/) {
	my ($tbl, $fld) = split('-', $dim);
	$tbl = $self->_dimensional_table( $tbl );
        return join('.', $tbl, $fld);
    }
    my $tbl = $self->_dimensional_table( $dim );
    my $fld = $dimension_fields{$dim};
    if (ref($fld) eq 'ARRAY') {
        my @flds = map { join('.', $tbl, $_) } @{$fld};
        if (($$fld[0] eq 'lo') && ($$fld[1] eq 'hi')) {
            return join( "||'-'||", @flds );
        }
        if (($$fld[0] eq 'local') && ($$fld[1] eq 'domain')) {
            return join( "||'\@'||", @flds );
        }
    }
    return join('.', $tbl, $fld);
}

###############################################################################
# INTERNAL METHOD.  Returns the cleaned table field name.
###############################################################################
sub _clean_field
{
    ###########################################################################
    #
    # XXX - The hyphen workaround is a hack that should
    #        be fixed at some point ... it was used to
    #        override the default field selection as
    #        defined in this module
    #
    ###########################################################################
    my ($dim) = @_;
    if ($dim =~ /\-/) {
        ($dim) = split('-', $dim);
    }
    return $dim;
}

###############################################################################
# INTERNAL METHOD. Pivots the result set (if necessary).  Returns the updated
# data set back to the caller.  If pivoting the data set would update the
# labels to be shown, they are updated automatically.
###############################################################################
sub _pivot
{
    my ($self, $results, $dototal) = @_;

    ###########################################################################
    # Get the dimension we're pivoting around.  If we don't need to pivot the
    # data, return the result set we were originally given.
    ###########################################################################
    my $cfg = $self->{'_report'};
    my $dim_pivot = $cfg->{'pivot'};
    return $results if (!$dim_pivot);

    ###########################################################################
    # Figure out which column number it is that we're pivoting around.
    ###########################################################################
    my ($fact) = keys %{$cfg->{'fact'}};
    my $dims   = $cfg->{'fact'}{$fact}{'dimension'};
    $dims = (ref($dims) eq 'ARRAY') ? $dims : [$dims];

    my $pivot_col = 0;
    my $found_pivot = undef;
    foreach my $dim (@{$dims})
    {
        if ($dim eq $dim_pivot)
        {
            $found_pivot = 1;
            last;
        }
        $pivot_col++;
    }
    die "Unable to determine column to pivot data around.\n" if (!$found_pivot);

    ###########################################################################
    # Pivot the data set, and return the pivoted data back to the caller.
    ###########################################################################
    $self->{'logger'}->log( "DEBUG", "Pivoting data around column $dim_pivot (col# $pivot_col)" );
    
    # pivot() changes the header array, and we don't really want that because
    # it will also change the object's private data members. create a dummy
    # @headers array that can be safely modified
    my @headers = ();
    
    # num of labels might be > num of colums in the table. restrict the num of
    # elements in @headers to match num of columns (needed for pivot)
    my $result_columns = scalar @{$results->[0]};
    foreach my $label (@{$self->{'_labels'}}) {
        push (@headers, $label);
        last if ( scalar(@headers) == $result_columns);
    }
    
    # also check if scalar @headers == $result_columns (needed for pivot)
    while (scalar @headers < $result_columns) {
        push (@headers, 'stupid string to fool pivot');
    }

    # run pivot() with the dummy @header array
    my @pivot = pivot( 'table'          => $results,
                       'headings'       => \@headers,
                       'pivot_column'   => $pivot_col,
                       ($dototal ? ('row_sum' => 'Total') : ()),
                     );
    return \@pivot;
}

###############################################################################
# INTERNAL METHOD. Unique-ifies a list, while keeping order intact.
###############################################################################
sub _uniqify_list
{
    my (@vals) = @_;
    my @results;
    my %seen;
    map { push( @results, $_ ) if (!$seen{$_}++) } @vals;
    return @results;
}

###############################################################################
# INTERNAL METHOD. Unique-ifies a list, while keeping order intact;
# (dumbly) detects placeholders and makes sure they are identical as well.
# this is tricky and annoying. if there's time in the future we should
# rewrite _generate_sql to take care of it for us, or arrange the data
# differently to make this cleaner.
###############################################################################
sub _uniqify_placeholder_lists {
    my ($list, $pholds) = @_;
    my $phi = $[;
    my (@nlist, @npholds, %seen) = ();
    my $i;
    for($i = $[; $i <= $#$list; $i++) {
        if(my $cnt = $list->[$i] =~ tr{?}{}) {
            croak "ran out of placeholder values"
                if(($#$pholds + $cnt) - 1 < $phi);

            my $phid = join('::', $list->[$i], @{$pholds}[$phi .. $phi + ($cnt - 1)]);
                
            if(!$seen{$phid}++) {
                push(@nlist, $list->[$i]);
                push(@npholds, @{$pholds}[$phi .. $phi + ($cnt - 1)]);
            }
            
            $phi += $cnt;
        } else {
            if(!$seen{$list->[$i]}++) {
                push(@nlist, $list->[$i]);
            }
        }
    }

    @$list = @nlist;
    @$pholds = @npholds;
}

###############################################################################
# INTERNAL METHOD.  Formats the results based on the provided options,
# returning the formatted results back to the caller.
###############################################################################
sub _format_results
{
    my ($self, $results, %opts) = @_;

    ###########################################################################
    # Figure out which format we're generating the report in.
    ###########################################################################
    my $format  = $opts{'format'} || $self->get_format();
    my $package = "PureMessage::Reports::Report::Format::$format";
    eval "require $package;";
    if ($@)
    {
        my $msg = "Unable to load report format '$format'";
        $self->{'logger'}->log( "ERROR", "$msg; $@" );
        die "$msg.\n";
    }

    ###########################################################################
    # Add some relevant report information to the options we pass through to
    # the formatter.
    ###########################################################################
    $opts{'labels'} = $self->get_labels();
    $opts{'legend'} = [$self->get_legend()];

    ###########################################################################
    # Instantiate the report formatter, and tell it to format the results.
    ###########################################################################
    my $formatter = $package->new();
    $results = $formatter->format( $results, %opts );
    $self->{'_mime_type'} = $formatter->mime_type();
    return $results;
}

###############################################################################
# INTERNAL METHOD.  Determines which reduction algorithm is in use by the given
# fact; "SUM" or "AVG".
###############################################################################
sub _fact_aggregate
{
    my ($fact) = @_;
    return undef unless $fact;

    # it's only runtime facts that are currently aggregated by averaging
    # all other facts are item counts so they get summarized
    return 'AVG' if $fact =~ /runtime/i;
    return 'SUM';
}

###############################################################################
# INTERNAL METHOD.  Expands a duration out to the number of seconds it
# represents.  Shamelessly stolen directly from ActiveState::Config::v_dur().
###############################################################################
my $float_rx = qr/(?:(?i)(?:[+-]?)(?:(?=[0123456789]|[.])(?:[0123456789]*)(?:(?:[.])(?:[0123456789]{0,}))?)(?:(?:[E])(?:(?:[+-]?)(?:[0123456789]+))|))/;
sub _expand_duration
{
    my ($v) = @_;
    my $tot = 0;
    my @parts = grep { length $_ } split m/($float_rx(?:)[smhdwy]?)/, $v;
    for my $part (@parts) {
        my ($n, $t) = ($part =~ /($float_rx)([smhdwy]?)/);
        die "invalid duration '$v'" unless defined $n;
        if    ($t eq 's' or !$t) { $tot += $n }
        elsif ($t eq 'm')        { $tot += $n * 60 }
        elsif ($t eq 'h')        { $tot += $n * 3600 }
        elsif ($t eq 'd')        { $tot += $n * 86400 }
        elsif ($t eq 'w')        { $tot += $n * 604800 }
        elsif ($t eq 'y')        { $tot += $n * 31536000 }
        else {
            die "invalid duration '$v'";
        }
    }
    return $tot;
}

1;
