package PureMessage::Reports::Report::Format::chart;

###############################################################################
# Required inclusions.
###############################################################################
use strict;
use Carp;
use Math::Round qw(nearest_ceil);
use base qw(PureMessage::Reports::Report::Format Exporter);

our @EXPORT_OK = qw(%defaults);

###############################################################################
# Defaults
###############################################################################
our %defaults = (
    'width'     => 600,
    'height'    => 250,
    );

###############################################################################
# Formats the report data as a chart.
###############################################################################
sub format
{
    my ($self, $data, %opts) = @_;

    ###########################################################################
    # Extract relevant options
    ###########################################################################
    $self->{'width'}    = $opts{'width'}    || $defaults{'width'};
    $self->{'height'}   = $opts{'height'}   || $defaults{'height'};

    ###########################################################################
    # Rotate the entire data set so that GD can make sense of it.  A data set
    # in the format of:
    #   [2003-03-01,    10,  100,   200],
    #   [2003-03-02,    15,  105,   205],
    # needs to be rotated to be in this format for GD:
    #   [2003-03-01, 2003-03-02]
    #   [        10,         15],
    #   [       100,        105],
    #   [       200,        205],
    ###########################################################################
    $self->{'data'} = [];           # empty data set
    $self->{'max_value'} = 0;       # blank min/max values
    $self->{'max_sum'} = 0;
    $self->{'min_value'} = 0;
    $self->{'min_sum'} = 0;
    $self->{'has_data'} = 0;        # have no data yet
    foreach my $row (@{$data})
    {
        $self->_inc_max_row_value( $row );
        $self->_add_row_rotated( $row );
        $self->{'has_data'} = 1;
    }

    ###########################################################################
    # Reset max values/sums so that we've got some spacing around the top edges
    # of the charts.
    ###########################################################################
    $self->{'max_value'} += $self->{'max_value'} * 0.05;
    $self->{'max_sum'}   += $self->{'max_sum'}   * 0.05;
    $self->{'max_value'} = nearest_ceil(1, $self->{'max_value'});
    $self->{'max_sum'}   = nearest_ceil(1, $self->{'max_sum'});
    $self->{'max_value'} = 1 if ($self->{'max_value'} < 1);
    $self->{'max_sum'}   = 1 if ($self->{'max_sum'}   < 1);

    ###########################################################################
    # Get an appropriate image to show; the chart, or a "no data" image.
    ###########################################################################
    my $image = '';
    if (!$self->{'has_data'})
    {
        # No data; read in the "no data" image.
        my $img_file = $self->no_data_image();

        open( FIN, "<$img_file" ) ||
            die "Failed to open the no data image $img_file; $!";
        binmode( FIN );
        my ($buf, $i) = ('', 0);
        while ($i = read( FIN, $buf, 1024 )) {
            $image .= $buf;
        }
        close( FIN );
    }
    else
    {
        # Have data; generate the chart image.
        my $chart = $self->chart();

        # Set numeric format based on how wide a data set we've got to look at.
        if ($self->{'max_value'} < 5) {
            $chart->set( 'y_number_format' => '%.2f' );
        }
        else {
            $chart->set( 'y_number_format' => '%d' );
        }
        
        if($opts{gdgraph}) {
            $chart->set( %{$opts{gdgraph}} );
        }

        # Plot the chart.
        my $plot = $chart->plot( $self->{'data'} );
        $image = $plot->png();
    }

    ###########################################################################
    # Return the formatted image back to the caller.
    ###########################################################################
    return $image;
}

###############################################################################
# INTERNAL METHOD.  Increments the max row value/sum.
###############################################################################
sub _inc_max_row_value
{
    my ($self, $val) = @_;
    my $key = shift @$val;
    my $s = 0;
    foreach my $n (@$val) {
        $self->{'max_value'} = $n if $n > $self->{'max_value'};
        $s += $n;
    }
    $self->{'max_sum'} = $s if $s > $self->{'max_sum'};
    unshift @$val, $key;
}

###############################################################################
# INTERNAL METHOD.  Adds the rotated row to our data set.
###############################################################################
sub _add_row_rotated
{
    my ($self, $row) = @_;

    # to rotate the data, push each row value on to the end of each of 
    # the rows.
    for (my $i = 0; $i < scalar @$row; $i++) {
        push(@{$self->{'data'}->[$i]}, $$row[$i]);
    }
}

###############################################################################
# Returns the MIME-Type of the formatted output.
###############################################################################
sub mime_type
{
    my ($self) = @_;
    if ($self->{'has_data'}) {
        return 'image/png';
    }
    return $self->mime_type_no_data_image();
}

###############################################################################
# Returns the name of the "no data" image to use.  In this base class this
# always returns a generic "no data" image, although it is recommended that
# derived chart classes provide their own image.
###############################################################################
sub no_data_image
{
    my ($self) = @_;
    my $imgname = 'img/no_data.gif';

    if ( exists $ENV{DOCUMENT_ROOT} ) {
	$imgname = $ENV{DOCUMENT_ROOT}.'/Reports/'.$imgname;
    }

    return $imgname;
}

###############################################################################
# Determines the MIME type of the "no data" image, returning it back to the
# caller.
###############################################################################
sub mime_type_no_data_image
{
    my ($self) = @_;
    my $img_file = $self->no_data_image();
    $img_file =~ s/.*\.//;
    return "image/$img_file";
}

###############################################################################
# Initializes the chart object and returns it back to the caller.  This method
# MUST be implemented by any derived charts.
###############################################################################
sub chart
{
    my ($self) = @_;
    croak "Chart interface not fully implemented in " . ref($self);
}

1;
