package PureMessage::Reports::RowBuffer;

# Provides functionality for in-memory buffering a set of rows for a given
# table, and methods for maintaining the buffer, including adding rows
# and flushing the buffer into the DBMS. Also provides table definitions
# for temporary buffering of data in the DBMS.

# Error handling: Routines that perform DBMS operations can die();
# they need to be evaled if we don't want the process to abort.
# Other errors are logged (if the logger is specified).

use strict;
use warnings;

sub new {
    my ($class, %opt) = @_;
    my $table_name = $opt{table_name};
    my $columns = $opt{columns};
    my $logger = $opt{logger};
    my %column;
    unless ($columns && $table_name) {
        die("unable to instantiate $class due to missing parameters");
    }

    # column array elements are in format column_name:column_type
    foreach my $col (@$columns) {
        my ($name, $type) = split ':', $col;
        unless ($type eq 'integer' || $type eq 'text') {
            my $error_msg = "Unknown type $type for column $name";
            if ($logger) {
                $logger->log("ERROR", $error_msg);
            }
            die $error_msg;
        }
        $column{$name} = $type;
    }

    my $buffer = [];
    my $self = { table_name => $table_name,
                 columns => \%column,
                 buffer => $buffer,
                 logger => $logger,
               };
    bless $self, $class;
    return $self;
}

sub _log {
    my $self = shift;
    if ($self->{logger}) {
        $self->{logger}->log(@_);
    }
}

# for strings: returns quoted, escaped value
# for numbers: returns unchanged value
# for undefs: returns string 'null'
# casts all values into the specified type
sub _sql_string_value {
    my ($self, $value, $type) = @_;
    $type = 'text' unless $type;

    if (defined $value) {
        if ($type eq 'text') {
            $value =~ s/'/''/g;
            $value =~ s/\\/\\\\/g;
            $value = "'$value'";
        }
    }
    else {
        $value = 'null';
    }
    return "${value}::$type";
}

# returns a single SQL statement composed as a union of single-row selects
# (one row for each element of the buffer), using the "insert into
# ... select * from (select ... union all select ...) union_view" syntax;
# returns undef if the buffer is empty
sub _get_flush_sql {
    my $self = shift;
    my @column = sort(keys %{$self->{columns}});
    my @select_row;
    foreach my $row (@{$self->{buffer}}) {
        my $select_row = 'select ';
        foreach my $column (@column) {
            $select_row .=
                $self->_sql_string_value($row->{$column},
                                         $self->{columns}->{$column}) . ', ';
        }
        if ($select_row) {
            $select_row =~ s/, \Z//;
            push @select_row, $select_row;
        }
    }
    my $union;
    if (scalar @select_row) {
        $union = "insert into " . $self->{table_name} .
                 " (" . join(', ', @column). ") select * from (" .
                 join(' union all ', @select_row) . ") union_view";
    }
    return $union;
}

# returns DDL statement that creates table that can hold buffered data
# in the DBMS
sub get_table_sql {
    my $self = shift;
    my $temp_modifier = shift;
    my $temp = $temp_modifier ? "temp" : "";
    my @column_name = sort(keys %{$self->{columns}});
    my @column;
    my $sql;
    foreach my $column_name (@column_name) {
        push @column, "$column_name $self->{columns}->{$column_name}";
    }
    if (scalar @column) {
        $sql = "create $temp table " . $self->{table_name} .
               " (" . join(', ', @column) . ")";
    }
    return $sql;
}

# stores the buffered rows into the database and empties the buffer;
# does not commit
# this routine can die() => needs to be evaled
sub flush {
    my ($self, $dbh) = @_;
    my $sql = $self->_get_flush_sql();
    if ($sql && $dbh) {
        $dbh->do($sql);
    }
    $self->{buffer} = [];
}

sub add_row {
    my $self = shift;
    my $row = shift;

    # validate the supplied column names
    foreach my $key (keys %$row) {
        unless (exists ${$self->{columns}}{$key}) {
            $self->_log("WARNING", "Ignoring unknown column $key");
            delete $row->{$key};
        }
    }
    push @{$self->{buffer}}, $row;
}

sub row_count {
    my $self = shift;
    return scalar @{$self->{buffer}};
}

return 1;
