#!perl

package PureMessage::Reports::TimeRanges;

use strict;
use warnings;
use Carp qw(croak carp);
use POSIX;
use Data::Dumper;
use HTTP::Date qw(str2time);
use Time::Local qw(timelocal);
use Math::Round qw(nlowmult);
use PureMessage::Reports::Util qw(round_time_hour round_time_day round_time_month);
use PureMessage::Reports::DB qw(get_dbh);
use ActiveState::DateTime qw(gmt_time_stamp);

# This module encapsulates functionality for insertion of the time period
# info into the PM reporting database, as well as reading the period info
# from the database and caching it into an in-memory structure.

# Error handling: routines in this module typically croak() on db errors.

return 1;

# one of the params can be the db handle
sub new {
    my $class = shift;
    my %args = (
        short           => 300,
        mid             => 3600,
        long            => 86400,
        hold_short      => (36 + 1) * 12,
        hold_mid        => 90 * 24,
        hold_long       => 2 * 365,
        @_
                );

    if(ref($args{data})) {
        %args = (%args, %{$args{data}});
        delete($args{data});
    }

    my $self = bless \%args, $class;
    $self->run unless($self->{no_autorun});
    $self;
}

# scheduler interface
sub run {
    shift->insert_period_info();
}

sub insert_period_info {
    my $self = shift;
    $self->{dbh} ||= get_dbh;

    my $now = $self->{'time-now'} || 'now';
    $now = gmt_time_stamp($now) if ($now ne "now"); # Append tz offset
    $now = "'$now'::timestamptz";

	$self->{dbh}->do(<<"__SQL__");
        --XXX This belongs in the schema.
	CREATE OR REPLACE FUNCTION generate_series(int, int)
	RETURNS SETOF int LANGUAGE plpgsql STRICT IMMUTABLE
	AS '
	DECLARE	start	ALIAS FOR \$1;
	    	stop 	ALIAS FOR \$2;
	    	i	int;
	BEGIN
	    FOR i IN start .. stop LOOP
		RETURN NEXT i;
	    END LOOP;
	    RETURN;
	END
	';

        -- Truncate $now to the most recent 5min time (e.g. 12:34:56 -> 12:30:00).
        SELECT date_trunc('sec',$now)
                - extract(epoch FROM (date_trunc('sec',$now) - date_trunc('hour',$now)))::int
                  % 300 * '1sec'::interval AS _now 
        INTO TEMP TABLE _now;

        LOCK TABLE prd_period IN EXCLUSIVE MODE;

        INSERT INTO prd_period(period_start, duration, period_end, super_period_start)
        SELECT X.*,  date_trunc('hour', X.period_start) AS super_period_start
        FROM (
            SELECT X.period_start, 
                   $self->{short} AS duration, 
                   X.period_start + '$self->{short} secs'::interval AS period_end
            FROM (
                SELECT  _now + '1hour'::interval - i *'$self->{short} secs'::interval AS period_start
                FROM    generate_series(0,$self->{hold_short} - 1) AS seq(i), _now
            ) AS X
                UNION ALL
            SELECT X.period_start, 
                   $self->{mid} AS duration, 
                   X.period_start + '$self->{mid} secs'::interval AS period_end
            FROM (
                SELECT  date_trunc('hour', _now) - i *'$self->{mid} secs'::interval AS period_start
                FROM    generate_series(0,$self->{hold_mid} - 1) AS seq(i), _now
            ) AS X
                UNION ALL
            SELECT X.period_start, 
                   $self->{long} AS duration, 
                   X.period_start + '$self->{long} secs'::interval AS period_end
            FROM (
                SELECT  date_trunc('day', _now) - i *'$self->{long} secs'::interval AS period_start
                FROM    generate_series(0,$self->{hold_long} - 1) AS seq(i), _now
            ) AS X
        ) AS X
        LEFT JOIN prd_period USING (period_start, period_end) WHERE id IS NULL;
__SQL__

    $self->{dbh}->commit();
}

# only for testing
sub delete_period_info {
    my $self = shift;
    $self->{dbh} ||= get_dbh;
    eval {
        $self->{dbh}->do("delete from prd_period");
        $self->{dbh}->commit();
    };
    if ($@) {
        croak "delete_period_info(): error while deleting period info";
    }
}

# Reads "short" period start times and corresponding db identifiers into
# an in-memory structure. Returns the count of cached periods.
sub cache_period_info {
    my $self = shift;
    $self->{dbh} ||= get_dbh;
    my $sql = <<END_OF_SQL;
select period_start, id from prd_period
where duration = ? order by period_start asc
END_OF_SQL
    my %period_cache;
    my ($period_start, $period_id);
    my $count = 0;
    if ($self->{period_cache_timestamp} and
        (time() - $self->{period_cache_timestamp} <= 30)) {

        # cached recently, don't do it again
        $count = scalar keys %{$self->{period_cache}};
    }
    else {
        eval {
            my $prepared_sql = $self->{dbh}->prepare($sql);
            $prepared_sql->execute($self->{short});
            $prepared_sql->bind_columns(\$period_start, \$period_id);
            while ($prepared_sql->fetch()) {
                $period_start = str2time($period_start);
                $period_cache{$period_start} = $period_id;
                $count++;
            }
        };
        if ($@) {
            croak "cache_period_info(): error while reading period info";
        }
        $self->{period_cache} = \%period_cache;
        $self->{period_cache_timestamp} = time();
    }
    return $count;
}

# Returns the database identifier for the given period start (note that
# we operate on shortest intervals only here). Input is expected in ISO
# string format. Returns undef if the identifier is not there.
sub get_period_id {
    my $self = shift;
    my $time = shift;
    my $cache_refreshed = 0;

    # map the specified time into the start of the shortest period
    # (numeric format)
    $time = str2time($time);
    unless ($time) {
        croak "get_period_id(): invalid input time format";
    }
    my $start_time = int($time / $self->{short}) * $self->{short};

    while (1) {
        if (exists($self->{period_cache}->{$start_time})) {
            return $self->{period_cache}->{$start_time};
        }
        last if $cache_refreshed;
        $self->cache_period_info();
        $cache_refreshed = 1;
    }
    return undef;
}
